import re, time
from .IxNetRestApi import IxNetRestApiException

class Traffic(object):
    def __init__(self, ixnObj=None):
        self.ixnObj = ixnObj

    def setMainObject(self, mainObject):
        # For Python Robot Framework support
        self.ixnObj = mainObject
        
    def configTrafficItem(self, mode=None, obj=None, trafficItem=None, endpoints=None, configElements=None):
        """
        Description
            Create or modify a Traffic Item.

            When creating a new Traffic Item, this API will return 3 object handles:
                 trafficItemObj, endpointSetObjList and configElementObjList

            NOTE:
                Each Traffic Item could create multiple endpoints and for each endpoint.
                you could provide a list of configElements for each endpoint.
                The endpoints and configElements must be in a list.

                - Each endpointSet allows you to configure the highLevelStream, which overrides configElements.
                - If you set bi-directional to True, then there will be two highLevelStreams that you could configure.
                - Including highLevelStream is optional.  Set highLevelStream to None to use configElements.

        Parameters
            mode: craete|modify

            obj: For "mode=modify" only. Provide the object to modify: trafficItemObj|configElementObj|endpointObj

            trafficItem: Traffic Item kwargs.

            endpoints: [list]: A list: [{name: sources:[], destionations:[], highLevelStreams: None, (add more endpoints)... ]
                               Scroll down to see example.

            configElements: [list]: Config Element kwargs.
                                    Each item in this list is aligned to the sequential order of your endpoint list.

        If mode is create:
            The required parameters are: mode, trafficItem, endpoints and configElements

        If mode is modify:
            The required parameters are: mode, obj, and one of the objects to modify (trafficIemObj, endpointObj or configElementObj).
            
            You need to provide the right object handle.

               To modify trafficItem:
                  Ex: /api/v1/sessions/1/ixnetwork/traffic/trafficItem/<id>

               To modify endpointSet:
                  Ex: /api/v1/sessions/1/ixnetwork/traffic/trafficItem/1/endpointSet/<id>

               To modify configElements = configElement object handlex
                  Ex: /api/v1/sessions/1/ixnetwork/traffic/trafficItem/1/configElement/<id>

               Look at sample script l2l3RestNgpy.py

        Traffic Item Parameters
            trafficType options:
               raw, ipv4, ipv4, ethernetVlan, frameRelay, atm, fcoe, fc, hdlc, ppp

            srcDestMesh:
               Defaults to one-to-one
               Options: manyToMany or fullMesh

            routeMesh:
               fullMesh or oneToOne

            allowSelfDestined: True or False

            trackBy: [list]: trackingenabled0, ethernetIiSourceaddress0, ethernetIiDestinationaddress0, ethernetIiPfcQueue0,
                             vlanVlanId0, vlanVlanUserPriority0, ipv4SourceIp0, sourceDestValuePair0, sourceDestEndpointPair0,
                             ipv4Precedence0, ipv4SourceIp0, flowGroup0, frameSize0 

        ConfigElement Parameters
            transmissionType:
               - continuous|fixedFrameCount|fixedDuration
               - custom (for burstPacketCount)

            frameCount: (For continuous and fixedFrameCount traffic)
            burstPacketCount: (For bursty traffic)
            frameSizeType: fixed|random
            frameSize: The packet size.
            
            frameRate: The rate to transmit packets
            frameRateType: bitsPerSecond|framesPerSecond|interPacketGap|percentLineRate
            duration: Set fixedDuration
            portDistribution: applyRateToAll|splitRateEvenly.  Default=applyRateToAll
            streamDistribution: splitRateEvenly|applyRateToAll. Default=splitRateEvently
            trackBy: <list>: Some options: flowGroup0, vlanVlanId0, ethernetIiDestinationaddress0, ethernetIiSourceaddress0,
                             sourcePort0, sourceDestPortPair0, ipv4DestIp0, ipv4SourceIp0, ipv4Precedence0,
                             ethernetIiPfcQueue0, frameSize0

        If frameSizeType == random
             incrementFrom: Frame size increment from.
             incrementTo: Frame size increment to.

        For bursty packet count,
              transmissionType = 'custom',
              burstPacketCount = 50000,

        Endpoints Parameters
            A list of topology, deviceGroup or protocol objects
                sources: Object in a list.
                destinations: Object in a lsit.

            Example:
               ['/api/v1/sessions/1/ixnetwork/topology/8']
               or a list ['.../topology/1', '.../topology/3']
               ['.../topology/1/deviceGroup/1', '.../topology/2/deviceGroup/1/ethernet/1/ipv4/1']


        USAGE EXAMPLE:
            To modify:
                trafficObj.configTrafficItem(mode='modify',
                                             obj='/api/v1/sessions/1/ixnetwork/traffic/trafficItem/1/configElement/1', 
                                             configElements={'transmissionType': 'continuous'})

                trafficObj.configTrafficItem(mode='modify',
                                             obj='/api/v1/sessions/1/ixnetwork/traffic/trafficItem/1',
                                             trafficItem={'trackBy': ['frameSize0', 'ipv4SourceIp0']})

            To create new Traffic Item:

            configTrafficItem(mode='create',
                              trafficItem = {
                                  'name':'Topo1 to Topo2',
                                  'trafficType':'ipv4',
                                  'biDirectional':True,
                                  'srcDestMesh':'one-to-one',
                                  'routeMesh':'oneToOne',
                                  'allowSelfDestined':False,
                                  'trackBy': ['flowGroup0', 'vlanVlanId0']},

                               endpoints = [{'name':'Flow-Group-1',
                                             'sources': [topologyObj1],
                                             'destinations': [topologyObj2],
                                             'highLevelStreamElements': None}],

                               configElements = [{'transmissionType': 'fixedFrameCount',
                                                  'frameCount': 50000,
                                                  'frameRate': 88,
                                                  'frameRateType': 'percentLineRate',
                                                  'frameSize': 128,
                                                  'portDistribution': 'applyRateToAll',
                                                  'streamDistribution': 'splitRateEvenly'
                                                  }]
            )

            To create a new Traffic Item and configure the highLevelStream:

            trafficObj.configTrafficItem(mode='create',
                                         trafficItem = {'name':'Topo3 to Topo4',
                                                       'trafficType':'ipv4',
                                                       'biDirectional':True,
                                                       'srcDestMesh':'one-to-one',
                                                       'routeMesh':'oneToOne',
                                                       'allowSelfDestined':False,
                                                       'trackBy': ['flowGroup0', 'vlanVlanId0']},
                                         endpoints = [{'name':'Flow-Group-1',
                                                        'sources': [topologyObj1],
                                                        'destinations': [topologyObj2],
                                                        'highLevelStreamElements': [
                                                           {
                                                               'transmissionType': 'fixedFrameCount',
                                                               'frameCount': 10000,
                                                               'frameRate': 18,
                                                               'frameRateType': 'percentLineRate',
                                                               'frameSize': 128}, 
                                                           {
                                                               'transmissionType': 'fixedFrameCount',
                                                               'frameCount': 20000,
                                                               'frameRate': 28,
                                                               'frameRateType': 'percentLineRate',
                                                               'frameSize': 228}
                                                         ]
                                                     }],
                                         configElements = None)
    

        Return: trafficItemObj, endpointSetObjList, configElementObjList
        """
        if mode == 'create':
            trafficItemUrl = self.ixnObj.sessionUrl+'/traffic/trafficItem'
        if mode == 'modify' and obj is None:
            raise IxNetRestApiException('Modifying Traffic Item requires a Traffic Item object')
        if mode == 'create' and trafficItem is None:
            raise IxNetRestApiException('Creating Traffic Item requires trafficItem kwargs')
        if mode == None:
            raise IxNetRestApiException('configTrafficItem Error: Must include mode: config or modify')

        # Don't configure config elements if user is configuring highLevelStreams
        isHighLevelStreamTrue = False

        # Create a new Traffic Item
        if mode == 'create' and trafficItem != None:
            if 'trackBy' in trafficItem:
                trackBy = trafficItem['trackBy']
                del trafficItem['trackBy']

            self.ixnObj.logInfo('configTrafficItem: %s : %s' % (trafficItemUrl, trafficItem), timestamp=False)
            response = self.ixnObj.post(trafficItemUrl, data=trafficItem)
            trafficItemObj = response.json()['links'][0]['href']

        if mode == 'modify' and trafficItem != None:
            trafficItemObj = obj
            if 'trackBy' in trafficItem:
                trackBy = trafficItem['trackBy']
                del trafficItem['trackBy']
            self.ixnObj.patch(self.ixnObj.httpHeader+trafficItemObj, data=trafficItem)

        # Create Endpoints
        if mode == 'create' and endpoints != None:
            if type(endpoints) != list:
                raise IxNetRestApiException('configTrafficItem error: Provide endpoints in a list')

            endpointSetObjList = []
            if 'trafficItemObj' not in locals():
                # Expect the user to pass in the endpoint object handle correctly and parse
                # out the traffic item object handle.
                trafficItemObj = self.ixnObj.sessionUrl.split('/endpointSet')[0]

            # endpoints = [{'name':'Flow-Group-1', 'sources': [topologyObj1], 'destinations': [topologyObj2], 'highLevelStreamElements': None}]
            for endPoint in endpoints:
                endpointSrcDst = {}
                # {'name':'Flow-Group-1', 'sources': [topologyObj1], 'destinations': [topologyObj2]}
                if 'name' in endPoint:
                    endpointSrcDst['name'] = endPoint['name']

                if 'sources' in endPoint:
                    endpointSrcDst['sources'] = endPoint['sources']
                
                if 'destinations' in endPoint:
                    endpointSrcDst['destinations'] = endPoint['destinations']

                if 'multicastDestinations' in endPoint:
                    endpointSrcDst['multicastDestinations'] = endPoint['multicastDestinations']

                if 'multicastReceivers' in endPoint:
                    endpointSrcDst['multicastReceivers'] = endPoint['multicastReceivers']

                if 'scalableDestinations' in endPoint:
                    endpointSrcDst['scalableDestinations'] = endPoint['scalableDestinations']

                if 'scalableSources' in endPoint:
                    endpointSrcDst['scalableSources'] = endPoint['scalableSources']

                response = self.ixnObj.post(self.ixnObj.httpHeader+trafficItemObj+'/endpointSet', data=endpointSrcDst)

                if 'highLevelStreamElements' in endPoint:
                    highLevelStream = endPoint['highLevelStreamElements']
                    # JSON doesn't support None.  In case user passed in {} instead of None.
                    if highLevelStream == {}:
                        highLevelStream = None
                else:
                    highLevelStream = None

                # Get the RETURNED endpointSet/# object
                endpointSetObj = response.json()['links'][0]['href']
                response = self.ixnObj.get(self.ixnObj.httpHeader+endpointSetObj)

                # This endpontSet ID is used for getting the corresponding Config Element ID
                # in case there are multiple endpoint sets created.
                endpointSetId = response.json()['id']
                endpointSetObjList.append(endpointSetObj)

                # An endpoint flow group could have two highLevelStream if bi-directional is enabled.x
                if highLevelStream != None:
                    isHighLevelStreamTrue = True
                    configElementObjList = None ;# Don't configure config elements if user is configuring highLevelStreams
                    streamNum = 1
                    for eachHighLevelStream in highLevelStream:
                        self.configConfigElements(self.ixnObj.httpHeader+trafficItemObj+'/highLevelStream/'+str(streamNum), eachHighLevelStream)
                        streamNum += 1

        if mode == 'modify' and endpoints != None:
            endpointSrcDst = {}
            if 'name' in endpoints:
                endpointSrcDst['name'] = endpoints['name']
            if 'sources' in endpoints:
                endpointSrcDst['sources'] = endpoints['sources']
            if 'destinations' in endpoints:
                endpointSrcDst['destinations'] = endpoints['destinations']

            endpointSetObj = obj
            self.ixnObj.patch(self.ixnObj.httpHeader+endpointSetObj, data=endpointSrcDst)

        if isHighLevelStreamTrue == False and configElements != None:
            if mode == 'create' and type(configElements) != list:
                raise IxNetRestApiException('configTrafficItem error: Provide configElements in a list')

            if mode == 'modify':
                configElementObj = obj
                self.configConfigElements(self.ixnObj.httpHeader+configElementObj, configElements)

            if mode == 'create':
                endpointResponse = self.ixnObj.get(self.ixnObj.httpHeader+trafficItemObj+'/endpointSet')

                index = 0
                configElementCounter = 1
                configElementObjList = []
                for eachEndpoint in endpointResponse.json():
                    configElementObj = trafficItemObj+'/configElement/'+str(configElementCounter)
                    configElementObjList.append(configElementObj)
                    self.configConfigElements(self.ixnObj.httpHeader+configElementObj, configElements[index])
                    if len(endpointSetObjList) == len(configElements):
                        index += 1
                    configElementCounter += 1

        if configElements is None:
            configElementObjList = []

        # Cannot configure tracking until endpoints are created. This is why
        # tracking config is at the end here.
        if mode in ['create', 'modify'] and 'trackBy' in locals():
            self.ixnObj.logInfo('Config Traffic Item statistic trackings', timestamp=False)
            self.ixnObj.patch(self.ixnObj.httpHeader+trafficItemObj+'/tracking', data={'trackBy': trackBy})

        # API server needs some time to complete processing the highlevel stream configuration before entering regenerate.
        if mode == 'create' and trafficItem != None:
            return [trafficItemObj, endpointSetObjList, configElementObjList]

    def configConfigElements(self, configElementObj, configElements):
        """
        Description
           Configure Traffic Item Config Elements. This function will collect all the ReST API's attributes
           and execute a PATCH in one single command instead of sending a a PATCH for each attribute.
           This avoids dependency breakage because some APIs require the type to be configured first.

           This function also handles high level stream configurations since the attributes are the same. 
           Pass in the highLevelStream obj for the parameter configElementObj

        Parameters
           configElementObj: <str:obj>: The config element object:
                             Ex: /api/v1/sessions/{1}/ixnetwork/traffic/trafficItem/{1}/configElement/{1}
                             highLevelStream obj Ex: /api/v1/sessions/{1}/ixnetwork/traffic/trafficItem/{1}/highLevelStream/{1}
        
           configElements: <dict>: This could also be highLevelStream elements.  
                           configElements = {'transmissionType': 'fixedFrameCount',
                                              'frameCount': 50000,
                                              'frameRate': 88,
                                              'frameRateType': 'percentLineRate',
                                              'frameSize': 128,
                                              'portDistribution': 'applyRateToAll',
                                              'streamDistribution': 'splitRateEvenly'
                                            }
           transmissionType:   fixedFrameCount|continuous|fixedDuration
           incrementFrom:      For frameSizeType = random.  Frame size from size. 
           incrementTo:        For frameSizeType = random. Frame size to size. 
           frameRateType:      percentLineRate|framesPerSecond
           portDistribution:   applyRateToAll|splitRateEvenly. Default=applyRateToAll
           streamDistribution: splitRateEvenly|applyRateToAll. Default=splitRateEvently
        """
        transmissionControlData = {}
        frameRateData = {}
        frameRateDistribution = {}
        for item in configElements.keys() :
            # These attributes are int type
            if item in ['burstPacketCount', 'duration', 'frameCount', 'interBurstGap', 'interStreamGap',
                        'iterationCount', 'minGapBytes', 'repeatBurst', 'startDelay']: 
                transmissionControlData.update({item: int(configElements[item])})

            if item in ['enableInterBurstGap', 'enableInterStreamGap','interBurstGapUnits',
                        'startDelayUnits', 'type']:
                transmissionControlData.update({item: str(configElements[item])})

            if item == 'frameRateType': 
                frameRateData.update({'type': str(configElements[item])})

            if item == 'frameRate': 
                frameRateData.update({'rate': float(configElements[item])})

            if item == 'portDistribution': 
                frameRateDistribution.update({'portDistribution': configElements[item]})
                
            if item == 'streamDistribution': 
                frameRateDistribution.update({'streamDistribution': configElements[item]})

        # Note: transmissionType is not an attribute in configElement. It is created to be more descriptive than 'type'.
        if 'transmissionType' in configElements:
            self.ixnObj.patch(configElementObj+'/transmissionControl', data={'type': configElements['transmissionType']})

        if transmissionControlData != {}:
            self.ixnObj.patch(configElementObj+'/transmissionControl', data=transmissionControlData)

        if frameRateData != {}:
            self.ixnObj.patch(configElementObj+'/frameRate', data=frameRateData)

        if 'frameSize' in configElements:
            self.ixnObj.patch(configElementObj+'/frameSize', data={'fixedSize': int(configElements['frameSize'])})

        if 'frameSizeType' in configElements:
            self.ixnObj.patch(configElementObj+'/frameSize', data={'type': configElements['frameSizeType']})
            if configElements['frameSizeType'] == 'random':
                self.ixnObj.patch(configElementObj+'/frameSize', data={'incrementFrom': configElements['incrementFrom'],
                                                                       'incrementTo': configElements['incrementTo']})
        if frameRateDistribution != {}:
            self.ixnObj.patch(configElementObj+'/frameRateDistribution', data=frameRateDistribution)        

    def getConfigElementObj(self, trafficItemObj=None, trafficItemName=None, endpointSetName=None):
        """
        Description
           Get the config element object handle.

           Use case #1: trafficItemName + endpointSetName
           Use case #2: trafficItemObj + endpointSetName

           Use case #3: trafficItemName only (Will assume there is only one configElement object which will be returned)
           Use case #4: trafficItemObj only  (Will assume there is only one configElement object which will be returned)

        Parameters
             
           trafficItemObj: <str obj>: The Traffic Item object.
           trafficItemName: <str>: The Traffic Item name.
           endpointSetName: <str>: The Traffic Item's EndpointSet name.

           How this works:
               - Users could create multiple EndpointSets within a Traffic Item.
               - Each EndpointSet has a unique object ID by default.
               - For each EndpointSet is created, a config element object is also created.
               - Each config element object handle is associated with an EndpointSet ID.
               - To be able to get the right config element object handle, we need to query
                 for the EndpointSet that you need for modifying.
               - Another terminology for EndpointSet is FlowGroup.
               - If you have multiple EndpointSets, you should give each EndpointSet a name
                 to make querying possible.
               - Otherwise, this function will assume there is only one EndpointSet created which will be returned.

        Usage examples:
           trafficObj.getConfigElementObj(trafficItemName='Raw MPLS/UDP', endpointSetName='EndpointSet-2')

           trafficObj.getConfigElementObj(trafficItemName='Raw MPLS/UDP', endpointSetName=None)

           trafficObj.getConfigElementObj(trafficItemObj='/api/v1/sessions/1/ixnetwork/traffic/trafficItem/1', 
                                          trafficItemName=None, endpointSetName=None)

           trafficObj.getConfigElementObj(trafficItemObj='/api/v1/sessions/1/ixnetwork/traffic/trafficItem/1', 
                                          trafficItemName=None, endpointSetName='EndpointSet-2')

        Return
           configElement: /api/v1/sessions/{id}/ixnetwork/traffic/trafficItem/{id}/configElement/{id}
        """
        if trafficItemObj:
            trafficItemName = self.getTrafficItemName(trafficItemObj)

        if trafficItemName:
            if endpointSetName:
                queryData = {'from': '/traffic',
                             'nodes': [{'node': 'trafficItem', 'properties': ['name'],
                                        'where': [{'property': 'name', 'regex': trafficItemName}]},
                                       {'node': 'endpointSet', 'properties': ['name'],
                                        'where': [{'property': 'name', 'regex': endpointSetName}]},
                      ]}

            if endpointSetName == None:
                queryData = {'from': '/traffic',
                             'nodes': [{'node': 'trafficItem', 'properties': ['name'],
                                        'where': [{'property': 'name', 'regex': trafficItemName}]},
                                       {'node': 'endpointSet', 'properties': [],
                                        'where': []},
                      ]}

            queryResponse = self.ixnObj.query(data=queryData)

            trafficItemList = queryResponse.json()['result'][0]['trafficItem']
            if trafficItemList == []:
                raise IxNetRestApiException('\nError: No traffic item name found: {0}'.format(trafficItemName))

            endpointSetList = queryResponse.json()['result'][0]['trafficItem'][0]['endpointSet']
            if endpointSetList == []:
                raise IxNetRestApiException('\nError: No endpointSet name: {0} found in Traffic Item name: {1}'.format(
                    endpointSetName, trafficItemName))

            endpointSetObj = queryResponse.json()['result'][0]['trafficItem'][0]['endpointSet'][0]['href']
            endpointSetId = endpointSetObj.split('/')[-1]

            # With the traffic item name and endpointSetId, get the Traffic Item's config element object handle.
            queryData = {'from': '/traffic',
                         'nodes': [{'node': 'trafficItem', 'properties': ['name'],
                                    'where': [{'property': 'name', 'regex': trafficItemName}]},
                                   {'node': 'configElement', 'properties': ['endpointSetId'],
                                    'where': [{'property': 'endpointSetId', 'regex': endpointSetId}]}
                  ]}

            queryResponse = self.ixnObj.query(data=queryData)
            configElementObj = queryResponse.json()['result'][0]['trafficItem'][0]['configElement'][0]['href']
            print(configElementObj)
            return configElementObj

    def getTransmissionType(self, configElement):
        # configElement: /api/v1/sessions/1/ixnetwork/traffic/trafficItem/1/configElement/1
        # Returns: fixedFrameCount, continuous

        response = self.ixnObj.get(self.ixnObj.httpHeader+configElement+'/transmissionControl')
        return response.json()['type']

    def configTrafficLatency(self, enabled=True, mode='storeForward'):
        # enabled = True|False
        # mode    = storeForward|cutThrough|forwardDelay|mef
        self.ixnObj.patch(self.ixnObj.sessionUrl+'/traffic/statistics/latency', data={'enabled':enabled, 'mode':mode})

    def showProtocolTemplates(self, configElementObj):
        """
        Description
           To show all the protocol template options. Mainly used for adding a protocol header
           to Traffic Item packets.

        Parameters
           configElementObj: /api/v1/sessions/1/ixnetwork/traffic/trafficItem/{id}/configElement/{id}
        """
        # Get a list of all the protocol templates:
        response = self.ixnObj.get(self.ixnObj.sessionUrl+'/traffic/protocolTemplate?skip=0&take=end')
        for eachProtocol in response.json()['data']:
            self.ixnObj.logInfo('%s: %s' % (eachProtocol['id'], eachProtocol['displayName']), timestamp=False)

    def showTrafficItemPacketStack(self, configElementObj):
        """
        Description
           Display a list of the current packet stack in a Traffic Item

           1: Ethernet II
           2: VLAN
           3: IPv4
           4: UDP
           5: Frame Check Sequence CRC-32

        Parameters
           configElementObj: /api/v1/sessions/1/ixnetwork/traffic/trafficItem/{id}/configElement/{id}
       """
        print()
        response = self.ixnObj.get(self.ixnObj.httpHeader+configElementObj+'/stack')
        self.ixnObj.logInfo('\n', timestamp=False)
        for (index, eachHeader) in enumerate(response.json()):
            self.ixnObj.logInfo('%s: %s' % (str(index+1), eachHeader['displayName']), timestamp=False)

    def addTrafficItemPacketStack(self, configElementObj, protocolStackNameToAdd, stackNumber, action='append'):
        """
        Description
           To either append or insert a protocol stack to an existing packet.

           You must know the exact name of the protocolTemplate to add by calling
           showProtocolTemplates() API and get the exact name  as a value for the parameter protocolStackNameToAdd.

           You must also know where to add the new packet header stack.  Use showTrafficItemPacketStack() to see
           your current stack numbers.

           This API returns the protocol stack object handle so you could use it to config its settings.

         Parameters
           configElementObj: /api/v1/sessions/1/ixnetwork/traffic/trafficItem/{id}/configElement/{id}

           action:
               append: To add after the specified stackNumber
               insert: To add before the specified stackNumber

           protocolStackNameToAdd: The name of the protocol stack to add.  To get a list of options,
                                   use API showProtocolTemplates().
                                   Some common ones: MPLS, IPv4, TCP, UDP, VLAN, IGMPv1, IGMPv2, DHCP, VXLAN

           stackNumber: The stack number to append or insert into.
                        Use showTrafficItemPacketStack() to view the packet header stack in order to know
                        which stack number to insert your new stack before or after the stack number.

        Example:
            addTrafficItemPacketStack(configElement, protocolStackNameToAdd='UDP',
                                      stackNumber=3, action='append', apiKey=apiKey, verifySslCert=False

        Returns:
            /api/v1/sessions/1/ixnetwork/traffic/trafficItem/{id}/configElement/{id}/stack/{id}
        """
        if action == 'append':
            action = 'appendprotocol'
        if action == 'insert':
            action = 'insertprotocol'

        # /api/v1/sessions/1
        match = re.match('http.*(/api.*sessions/[0-9]).*', self.ixnObj.sessionUrl)
        if match:
            apiHeader = match.group(1)

        # /api/v1/sessions/1/ixnetwork/traffic/trafficItem/1/configElement/1
        arg1 = configElementObj+'/stack/' + str(stackNumber)

        # Display a list of the current packet stack
        response = self.ixnObj.get(self.ixnObj.httpHeader+configElementObj+'/stack')
        for (index, eachHeader) in enumerate(response.json()):
            self.ixnObj.logInfo('{0}: {1}'.format(index+1, eachHeader['displayName']), timestamp=False)

        # Get a list of all the protocol templates:
        response = self.ixnObj.get(self.ixnObj.sessionUrl+'/traffic/protocolTemplate?skip=0&take=end')

        protocolTemplateId = None
        for eachProtocol in response.json()['data']:
            if bool(re.match('^%s$' % protocolStackNameToAdd, eachProtocol['displayName'].strip(), re.I)):
                # /api/v1/sessions/1/traffic/protocolTemplate/30
                protocolTemplateId =  eachProtocol['links'][0]['href']

        if protocolTemplateId == None:
            raise IxNetRestApiException('No such protocolTemplate name found: {0}'.format(protocolStackNameToAdd))
        self.ixnObj.logInfo('protocolTemplateId: %s' % protocolTemplateId, timestamp=False)
        data = {'arg1': arg1, 'arg2': protocolTemplateId}
        response = self.ixnObj.post(self.ixnObj.httpHeader+configElementObj+'/stack/operations/%s' % action, data=data)

        self.ixnObj.waitForComplete(response, self.ixnObj.httpHeader+configElementObj+'/stack/operations/appendprotocol/'+response.json()['id'])

        # /api/v1/sessions/1/ixnetwork/traffic/trafficItem/1/configElement/1/stack/4
        self.ixnObj.logInfo('addTrafficItemPacketStack: Returning: %s' % response.json()['result'], timestamp=False)
        return response.json()['result']

    def getTrafficItemPktHeaderStackObj(self, configElementObj=None, trafficItemName=None, packetHeaderName=None):
        """
        Description
           Get the Traffic Item packet header stack object based on the displayName.
           To get the displayName, you could call this function: self.showTrafficItemPacketStack(configElementObj)

           For this function, you could either pass in a configElement object or the Traffic Item name.
           
        Parameters
           configElementObj: <str>: Optional: The configElement object.
                             Example: /api/v1/sessions/1/ixnetwork/traffic/trafficItem/1/configElement/1

           trafficItemName: <str>: Optional: The Traffic Item name.

           packetHeaderName: <str>: Mandatory: The packet header name.
                             Example: ethernet II, mpls, ipv4, ...

        Return
           The stack object
        """
        if configElementObj == None:
            # Expect user to pass in the Traffic Item name if user did not pass in a configElement object.
            queryData = {'from': '/traffic',
                         'nodes': [{'node': 'trafficItem', 'properties': ['name'],
                                    'where': [{'property': 'name', 'regex': trafficItemName}]}]}

            queryResponse = self.ixnObj.query(data=queryData)

            if queryResponse.json()['result'][0]['trafficItem'] == []:
                raise IxNetRestApiException('\nNo such Traffic Item name found: %s' % trafficItemName)
                
            trafficItemObj = queryResponse.json()['result'][0]['trafficItem'][0]['href']
            configElementObj = trafficItemObj+'/configElement/1'
        
        response = self.ixnObj.get(self.ixnObj.httpHeader+configElementObj+'/stack')

        for eachStack in response.json():
            currentStackDisplayName = eachStack['displayName'].strip()
            self.ixnObj.logInfo('Packet header name: {0}'.format(currentStackDisplayName), timestamp=False)
            if bool(re.match('^{0}$'.format(packetHeaderName), currentStackDisplayName, re.I)):
                self.ixnObj.logInfo('\nstack: {0}: {1}'.format(eachStack, currentStackDisplayName), timestamp=False)
                stackObj = eachStack['links'][0]['href']
                break
        else:
            raise IxNetRestApiException('\nError: No such stack name found. Verify stack name existence and spelling: %s' % packetHeaderName)

        return stackObj

    def showTrafficItemStackLink(self, configElementObj):
        # Return a list of configured Traffic Item packet header in sequential order.
        #   1: Ethernet II
        #   2: MPLS
        #   3: MPLS
        #   4: MPLS
        #   5: MPLS
        #   6: IPv4
        #   7: UDP
        #   8: Frame Check Sequence CRC-32

        stackList = []
        response = self.ixnObj.get(self.ixnObj.httpHeader+configElementObj+'/stackLink')
        self.ixnObj.logInfo('\n', timestamp=False)
        for eachStackLink in response.json():
            if eachStackLink['linkedTo'] != 'null':
                self.ixnObj.logInfo(eachStackLink['linkedTo'], timestamp=False)
                stackList.append(eachStackLink['linkedTo'])
        return stackList

    def getPacketHeaderStackIdObj(self, configElementObj, stackId):
        """
        Desciption
           This API should be called after calling showTrafficItemPacketStack(configElementObj) in
           order to know the stack ID number to use.  Such as ...

            Stack1: Ethernet II
            Stack2: MPLS
            Stack3: MPLS
            Stack4: MPLS
            Stack5: MPLS
            Stack6: IPv4
            Stack7: UDP
            Stack8: Frame Check Sequence CRC-32

        Parameters
           configElementObj: /api/v1/sessions/1/ixnetwork/traffic/trafficItem/{id}/configElement/{id}
           stackId: In this example, IPv4 stack ID is 6.

         Return stack ID object: /api/v1/sessions/1/ixnetwork/traffic/trafficItem/{id}/configElement/{id}/stack/{id}
        """
        response = self.ixnObj.get(self.ixnObj.httpHeader+configElementObj+'/stack')
        self.ixnObj.logInfo('\n', timestamp=False)
        for (index, eachHeader) in enumerate(response.json()):
            self.ixnObj.logInfo('{0}: {1}'.format(index+1, eachHeader['displayName']), timestamp=False)
            if stackId == index+1:
                self.ixnObj.logInfo('\tReturning: %s' % self.ixnObj.httpHeader+eachHeader['links'][0]['href'], timestamp=False)
                return eachHeader['links'][0]['href']

    def modifyTrafficItemPacketHeader(self, configElementObj, packetHeaderName, fieldName, values):
        """
        Description
           Modify any Traffic Item packet header.  You will need to use the IxNetwor API browser
           to understand the packetHeaderName, fieldName and data to modify.

           Since a Traffic Item could contain many endpointSet (Flow Groups), a Traffic Item could 
           have multiple configElement objects.  A configElementObj is the object handle for an 
           endpointSet.  You have to get the configElement object first.  To get the ConfigElement
           object, you call getConfigElementObj().  

           self.getConfigElementObj(self, trafficItemObj=None, trafficItemName=None, endpointSetName=None):
           Use case #1: trafficItemName + endpointSetName
           Use case #2: trafficItemObj + endpointSetName
           Use case #3: trafficItemName only (Will assume there is only one configElement object which will be returned)
           Use case #4: trafficItemObj only  (Will assume there is only one configElement object which will be returned)

        Parameters
           configElementObj: <str|obj>: The Traffic Item's Config Element object handle.
           packetHeaderName: <str>: The packet header name. You could get the list of names from the 
                                    IxNetwork API browser under trafficItem/{id}/configElement/{id}/stack.
           fieldName: <str>: The packet header field name. View API browser under:
                             trafficItem/{id}/configElement/{id}/stack/{id}/field
           values: <dict>: Any amount of attributes from the /stack/{id}/field/{id} to modify.

        Example:  For IP Precedence TOS 
           packetHeaderName='ipv4'
           fieldName='Precedence'
           values={'fieldValue': '011 Flash'}
        """
        # /api/v1/sessions/1/ixnetwork/traffic/trafficItem/1/configElement/1/stack/6
        stackIdObj = self.getTrafficItemPktHeaderStackObj(configElementObj=configElementObj,
                                                          packetHeaderName=packetHeaderName)
        
        self.configPacketHeaderField(stackIdObj, fieldName, values)

    def modifyTrafficItemIpPriorityTos(self, trafficItemObj=None, trafficItemName=None, endpointSetName=None,
                                       packetHeaderName='ipv4', fieldName='Precedence', values=None):
        """
        Description
           Modify a Traffic Item Flow group IP Priority TOS fields.

        Parameters
           value: <dict>: {'fieldValue': '000 Routine'|'001 Priority'|'010 Immediate'|'011 Flash'|'100 Flash Override'
                           '101 CRITIC/ECP'|'110 Internetwork Control'}
        
           trafficItemObj: <str|obj>: The Traffic Item object handle.
           trafficItemName: <str|obj>: The Traffic Item name.
           endpointSetName: <str|obj>: The endpointSet name (Flow-Group).

           Option #1: trafficItemName + endpointSetName
           Option #2: trafficItemObj + endpointSetName
           Option #3: trafficItemName only (Will assume there is only one configElement object)
           Option #4: trafficItemObj only  (Will assume there is only one configElement object)

        Requirement
           Call self.getConfigElementObj() to get the config element object first.

        Example
           trafficObj.modifyTrafficItemIpPriorityTos(trafficItemName='Raw MPLS/UDP', values={'fieldValue': '001 Priority'})
        """
        configElementObj = self.getConfigElementObj(trafficItemObj=trafficItemObj, trafficItemName=trafficItemName,
                                                    endpointSetName=endpointSetName)

        self.modifyTrafficItemPacketHeader(configElementObj, packetHeaderName=packetHeaderName,
                                           fieldName=fieldName, values=values)

    def modifyTrafficItemDestMacAddress(self, trafficItemObj=None, trafficItemName=None, endpointSetName=None, values=None):
        """
        Description
           Modify a Traffic Item Flow group IP Priority TOS fields.

        Parameters
           value: <'str'|dict>: 
                  If str: The mac address address
                  If dict: Any or all the properties and the values:

                          {'valueType': 'increment',
                           'startValue': destMacAddress,
                           'stepValue': '00:00:00:00:00:00',
                           'countValue': 1,
                           'auto': False}
           
           trafficItemObj: <str|obj>: The Traffic Item object handle.
           trafficItemName: <str|obj>: The Traffic Item name.
           endpointSetName: <str|obj>: The endpointSet name (Flow-Group).

           Option #1: trafficItemName + endpointSetName
           Option #2: trafficItemObj + endpointSetName
           Option #3: trafficItemName only (Will assume there is only one configElement object)
           Option #4: trafficItemObj only  (Will assume there is only one configElement object)

        Requirement
           Call self.getConfigElementObj() to get the config element object first.

        Example
           trafficObj.modifyTrafficItemDestMacAddress(trafficItemName='Raw MPLS/UDP', values='00:01:01:02:00:01')
        """
        
        if type(values) == str:
            values = {'valueType': 'increment',
                      'startValue': values,
                      'stepValue': '00:00:00:00:00:00',
                      'countValue': 1,
                      'auto': False}

        configElementObj = self.getConfigElementObj(trafficItemObj=trafficItemObj, trafficItemName=trafficItemName,
                                                    endpointSetName=endpointSetName)

        self.modifyTrafficItemPacketHeader(configElementObj, packetHeaderName='ethernet',
                                           fieldName='Destination MAC Address', values=values)

    def showPacketHeaderFieldNames(self, stackObj):
        """
        Description
           Get all the packet header field names.

        Parameters
           stackObj = /api/v1/sessions/1/ixnetwork/traffic/trafficItem/{id}/configElement/{id}/stack/{id}

        Example for Ethernet stack field names
           1: Destination MAC Address
           2: Source MAC Address
           3: Ethernet-Type
           4: PFC Queue
        """
        self.ixnObj.logInfo('showPacketHeaderFieldNames: %s' % stackObj+'/field', timestamp=False)
        response = self.ixnObj.get(self.ixnObj.httpHeader+stackObj+'/field')
        for eachField in  response.json():
            id = eachField['id']
            fieldName = eachField['displayName']
            self.ixnObj.logInfo('\t{0}: {1}'.format(id, fieldName), timestamp=False)

    def convertTrafficItemToRaw(self, trafficItemName):
        """
        Description

        Parameter
        """
        trafficItemObj = self.getTrafficItemObjByName(trafficItemName)
        if trafficItemObj == 0:
            raise IxNetRestApiException('\nNo such Traffic Item name: %s' % trafficItemName)
        self.ixnObj.post(self.ixnObj.sessionUrl+'/traffic/trafficItem/operations/converttoraw', data={'arg1': trafficItemObj})

    def configPacketHeaderField(self, stackIdObj, fieldName, data):
        """
        Desciption
            Configure raw packets in a Traffic Item.
            In order to know the field names to modify, use showPacketHeaderFieldNames() to display the names:

        stackIdObj: /api/v1/sessions/1/ixnetwork/traffic/trafficItem/{id}/configElement/{id}/stack/{id}

        fieldName: The field name in the packet header to modify.
                   Example. In a MPLS packet header, the fields would be "Label value", "MPLS Exp", etc.

                   Note: Use showPacketHeaderFieldNames(stackObj) API to dispaly your options.

        data: Example:
             data={'valueType': 'valueList', 'valueList': ['1001', '1002'], auto': False}
             data={'valueType': 'increment', 'startValue': '1.1.1.1', 'stepValue': '0.0.0.1', 'countValue': 2}
             data={'valueType': 'increment', 'startValue': '00:01:01:01:00:01', 'stepValue': '00:00:00:00:00:01'}
             data={'valueType': 'increment', 'startValue': 1001, 'stepValue': 1, 'countValue': 2, 'auto': False}
        
        Example: To modify MPLS field:
            packetHeaderObj = trafficObj.getTrafficItemPktHeaderStackObj(trafficItemName='Raw MPLS/UDP', packetHeaderName='mpls')
            trafficObj.configPacketHeaderField(packetHeaderObj,
                                               fieldName='MPLS Exp',
                                               data={'valueType': 'increment',
                                               'startValue': '4',
                                               'stepValue': '1',
                                               'countValue': 1,
                                               'auto': False})
        """
        fieldId = None
        # Get the field ID object by the user defined fieldName
        response = self.ixnObj.get(self.ixnObj.httpHeader+stackIdObj+'/field')
        for eachFieldId in response.json():
            if bool(re.match(fieldName, eachFieldId['displayName'], re.I)):
                fieldId = eachFieldId['id']

        if fieldId == None:
            raise IxNetRestApiException('Failed to located your provided fieldName:', fieldName)

        self.ixnObj.logInfo('configPacketHeaderFieldId:  fieldIdObj: %s' % stackIdObj+'/field/'+str(fieldId), timestamp=False)
        response = self.ixnObj.patch(self.ixnObj.httpHeader+stackIdObj+'/field/'+str(fieldId), data=data)

    def getPacketHeaderAttributesAndValues(self, streamObj, packetHeaderName, fieldName):
        """
        Parameters
           streamObj: <str>: configElementObj|highLevelStreamObj
                      Ex: /api/v1/sessions/{id}/ixnetwork/traffic/trafficItem/{1}/configElement/{id} 
                       or /api/v1/sessions/{id}/ixnetwork/traffic/trafficItem/{1}/highLevelStream/{id}

           packetHeaderName: <str>: Display Name of the stack.  Example: Ethernet II, VLAN, IPv4, TCP, etc. 

           fieldName: <str>: Display Name of the field.  Example: If packetHeaderName is Ethernet II, field names could be
                                    Destination MAC Address, Source MAC Address, Ethernet-Type and PFC Queue.
                                    You will have to know these field names. To view them, make your configurations
                                    and then go on the API browser and go to:

                            /api/v1/sessions/{id}/ixnetwork/traffic/trafficItem/{1}/configElement/{id}/stack/{id}/field

        Example:
            streamObj = '/api/v1/sessions/{id}/ixnetwork/traffic/trafficItem/{id}/configElement/{id}'
            data= trafficObj.getPacketHeaderAttributesAndValues(streamObj,
                                                                'Ethernet II', 'Source MAC Address')
            data['singleValue'] = For single value.
            data['startValue'] = If it is configured to increment.

        Returns
            All the attributes and values in JSON format.
        """
        response = self.ixnObj.get(self.ixnObj.httpHeader+streamObj+'/stack')
        for eachStack in response.json():
            stackHref = eachStack['links'][0]['href']
            response = self.ixnObj.get(self.ixnObj.httpHeader+stackHref)
            #  need to do strip() to response.json()['displayName'] because some displayName has trailing spaces. 
            #  for example:  "IPv4 " 
            if packetHeaderName == response.json()['displayName'].strip():
                response = self.ixnObj.get(self.ixnObj.httpHeader+stackHref+'/field')
                for eachField in response.json():
                    if fieldName == eachField['displayName']:
                        return eachField

    def configEgressCustomTracking(self, trafficItemObj, offsetBits, widthBits):
        """
        Description
           Configuring custom egress tracking. User must know the offset and the bits width to track.
           In most use cases, packets ingressing the DUT gets modified by the DUT and to track the
           correctness of the DUT's packet modification, use this API to verify the receiving port's packet
           offset and bit width.
        """
        # Safety check: Apply traffic or else configuring egress tracking won't work.
        self.applyTraffic()
        self.ixnObj.patch(self.ixnObj.httpHeader+trafficItemObj+'/tracking/egress',
                   data={'encapsulation': 'Any: Use Custom Settings',
                         'customOffsetBits': offsetBits,
                         'customWidthBits': widthBits
                     })
        self.ixnObj.patch(self.ixnObj.httpHeader+trafficItemObj, data={'egressEnabled': True})
        self.regenerateTrafficItems()
        self.applyTraffic()

    def createEgressStatView(self, trafficItemObj, egressTrackingPort, offsetBit, bitWidth,
                             egressStatViewName='EgressStatView', ingressTrackingFilterName=None):
        """
        Description
           Create egress statistic view for egress stats.

        """
        egressTrackingOffsetFilter = 'Custom: ({0}bits at offset {1})'.format(int(bitWidth), int(offsetBit))
        trafficItemName = self.getTrafficItemName(trafficItemObj)

        # Get EgressStats
        # Create Egress Stats
        self.ixnObj.logInfo('Creating new statview for egress stats...')
        response = self.ixnObj.post(self.ixnObj.sessionUrl+'/statistics/view',
                             data={'caption': egressStatViewName,
                                   'treeViewNodeName': 'Egress Custom Views',
                                   'type': 'layer23TrafficFlow',
                                   'visible': True})

        egressStatView = response.json()['links'][0]['href']
        self.ixnObj.logInfo('egressStatView Object: %s' % egressStatView)
        # /api/v1/sessions/1/ixnetwork/statistics/view/12

        self.ixnObj.logInfo('Creating layer23TrafficFlowFilter')
        # Dynamically get the PortFilterId
        response = self.ixnObj.get(self.ixnObj.httpHeader+egressStatView+'/availablePortFilter')
        portFilterId = []
        for eachPortFilterId in response.json():
            #192.168.70.10/Card2/Port1
            self.ixnObj.logInfo('\tAvailable PortFilterId: %s' % eachPortFilterId['name'], timestamp=False)
            if eachPortFilterId['name'] == egressTrackingPort:
                self.ixnObj.logInfo('\tLocated egressTrackingPort: %s' % egressTrackingPort, timestamp=False)
                portFilterId.append(eachPortFilterId['links'][0]['href'])
                break
        if portFilterId == []:
            raise IxNetRestApiException('No port filter ID found')
        self.ixnObj.logInfo('PortFilterId: %s' % portFilterId)

        # Dynamically get the Traffic Item Filter ID
        response = self.ixnObj.get(self.ixnObj.httpHeader+egressStatView+'/availableTrafficItemFilter')
        availableTrafficItemFilterId = []
        for eachTrafficItemFilterId in response.json():
            if eachTrafficItemFilterId['name'] == trafficItemName:
                availableTrafficItemFilterId.append(eachTrafficItemFilterId['links'][0]['href'])
                break
        if availableTrafficItemFilterId == []:
            raise IxNetRestApiException('No traffic item filter ID found.')

        self.ixnObj.logInfo('availableTrafficItemFilterId: %s' % availableTrafficItemFilterId, timestamp=False)
        # /api/v1/sessions/1/ixnetwork/statistics/view/12
        self.ixnObj.logInfo('egressStatView: %s' % egressStatView, timestamp=False)
        layer23TrafficFlowFilter = self.ixnObj.httpHeader+egressStatView+'/layer23TrafficFlowFilter'
        self.ixnObj.logInfo('layer23TrafficFlowFilter: %s' % layer23TrafficFlowFilter, timestamp=False)
        response = self.ixnObj.patch(layer23TrafficFlowFilter,
                              data={'egressLatencyBinDisplayOption': 'showEgressRows',
                                    'trafficItemFilterId': availableTrafficItemFilterId[0],
                                    'portFilterIds': portFilterId,
                                    'trafficItemFilterIds': availableTrafficItemFilterId})

        # Get the egress tracking filter
        egressTrackingFilter = None
        ingressTrackingFilter = None
        response = self.ixnObj.get(self.ixnObj.httpHeader+egressStatView+'/availableTrackingFilter')
        self.ixnObj.logInfo('Available tracking filters for both ingress and egress...', timestamp=False)
        for eachTrackingFilter in response.json():
            self.ixnObj.logInfo('\tFilter Name: {0}: {1}'.format(eachTrackingFilter['id'], eachTrackingFilter['name']), timestamp=False)
            if bool(re.match('Custom: *\([0-9]+ bits at offset [0-9]+\)', eachTrackingFilter['name'])):
                egressTrackingFilter = eachTrackingFilter['links'][0]['href']

            if ingressTrackingFilterName is not None:
                if eachTrackingFilter['name'] == ingressTrackingFilterName:
                    ingressTrackingFilter = eachTrackingFilter['links'][0]['href']

        if egressTrackingFilter is None:
            raise IxNetRestApiException('Failed to locate your defined custom offsets: {0}'.format(egressTrackingOffsetFilter))

        # /api/v1/sessions/1/ixnetwork/statistics/view/23/availableTrackingFilter/3
        self.ixnObj.logInfo('Located egressTrackingFilter: %s' % egressTrackingFilter, timestamp=False)
        enumerationFilter = layer23TrafficFlowFilter+'/enumerationFilter'
        response = self.ixnObj.post(enumerationFilter,
                             data={'sortDirection': 'ascending',
                                   'trackingFilterId': egressTrackingFilter})

        if ingressTrackingFilterName is not None:
            self.ixnObj.logInfo('Located ingressTrackingFilter: %s' % egressTrackingFilter, timestamp=False)
            response = self.ixnObj.post(enumerationFilter,
                                 data={'sortDirection': 'ascending',
                                       'trackingFilterId': ingressTrackingFilter})

        # Must enable one or more egress statistic counters in order to enable the
        # egress tracking stat view object next.
        #   Enabling: ::ixNet::OBJ-/statistics/view:"EgressStats"/statistic:"Tx Frames"
        #   Enabling: ::ixNet::OBJ-/statistics/view:"EgressStats"/statistic:"Rx Frames"
        #   Enabling: ::ixNet::OBJ-/statistics/view:"EgressStats"/statistic:"Frames Delta"
        #   Enabling: ::ixNet::OBJ-/statistics/view:"EgressStats"/statistic:"Loss %"
        response = self.ixnObj.get(self.ixnObj.httpHeader+egressStatView+'/statistic')
        for eachEgressStatCounter in response.json():
            eachStatCounterObject = eachEgressStatCounter['links'][0]['href']
            eachStatCounterName = eachEgressStatCounter['caption']
            self.ixnObj.logInfo('\tEnabling egress stat counter: %s' % eachStatCounterName, timestamp=False)
            self.ixnObj.patch(self.ixnObj.httpHeader+eachStatCounterObject, data={'enabled': True})

        self.ixnObj.patch(self.ixnObj.httpHeader+egressStatView, data={'enabled': True})
        self.ixnObj.logInfo('createEgressCustomStatView: Done')

        return egressStatView

    def enableTrafficItem(self, trafficItemNumber):
        url = self.ixnObj.sessionUrl+'/traffic/trafficItem/%s' % str(trafficItemNumber)
        response = self.ixnObj.patch(url, data={"enabled": "true"})

    def disableTrafficItem(self, trafficItemNumber):
        url = self.ixnObj.sessionUrl+'/traffic/trafficItem/%s' % str(trafficItemNumber)
        response = self.ixnObj.patch(url, data={"enabled": "false"})

    def enableAllTrafficItems(self, mode=True):
        """
        Description
           Enable or Disable all Traffic Items.
         
        Parameter
           mode: True|False
                 True: Enable all Traffic Items
                 False: Disable all Traffic Items
        """
        queryData = {'from': '/traffic',
                     'nodes': [{'node': 'trafficItem', 'properties': [], 'where': []}]}
        queryResponse = self.ixnObj.query(data=queryData)
        for trafficItem in queryResponse.json()['result'][0]['trafficItem']:
            self.ixnObj.patch(self.ixnObj.httpHeader + trafficItem['href'], data={'enabled': mode})

    def isTrafficItemNameExists(self, trafficItemName):
        """
        Description
           Verify if the Traffic Item name exists in the configuration.

        Parameter
           trafficItemName: The Traffic Item name to verify
        """
        trafficItemNameExists = False
        response = self.ixnObj.get(self.ixnObj.sessionUrl+'/traffic/trafficItem')
        for eachTrafficItem in response.json():
            if eachTrafficItem['name'] == trafficItemName:
                return True
        return False

    def enablePacketLossDuration(self):
        self.ixnObj.patch(self.ixnObj.sessionUrl+'/traffic/statistics/packetLossDuration', data={'enabled': 'true'})

    def disablePacketLossDuration(self):
        self.ixnObj.patch(self.ixnObj.sessionUrl+'/traffic/statistics/packetLossDuration', data={'enabled': 'false'})

    def checkTrafficState(self, expectedState=['stopped'], timeout=60, ignoreException=False):
        """
        Description
            Check the traffic state for the expected state.
            This is best used to verify that traffic has started before calling getting stats.

        Traffic states are:
            startedWaitingForStats, startedWaitingForStreams, started, stopped,
            stoppedWaitingForStats, txStopWatchExpected, locked, unapplied

        Parameters
            expectedState: <str>:  Input a list of expected traffic state.
                            Example: ['started', startedWaitingForStats'] <-- This will wait until stats has arrived.

            timeout: <int>: The amount of seconds you want to wait for the expected traffic state.
                      Defaults to 45 seconds.
                      In a situation where you have more than 10 pages of stats, you will
                      need to increase the timeout time.

            ignoreException: <bool>: If True, return 1 as failed, and don't raise an Exception.

        Return
            1: If failed.
        """
        if type(expectedState) != list:
            expectedState.split(' ')

        self.ixnObj.logInfo('checkTrafficState: Expecting state: {0}\n'.format(expectedState))
        for counter in range(1,timeout+1):
            response = self.ixnObj.get(self.ixnObj.sessionUrl+'/traffic', silentMode=True)
            currentTrafficState = response.json()['state']
            if currentTrafficState == 'unapplied':
                self.ixnObj.logWarning('\nCheckTrafficState: Traffic is UNAPPLIED')
                self.applyTraffic()

            self.ixnObj.logInfo('\ncheckTrafficState: {trafficState}: Expecting: {expectedStates}.'.format(trafficState=currentTrafficState,
                                                                                                           expectedStates=expectedState),
                                timestamp=False)
            self.ixnObj.logInfo('\tWaited {counter}/{timeout} seconds'.format(counter=counter, timeout=timeout), timestamp=False)
            
            if counter <= timeout and currentTrafficState not in expectedState:
                time.sleep(1)
                continue

            if counter <= timeout and currentTrafficState in expectedState:
                time.sleep(8)
                self.ixnObj.logInfo('checkTrafficState: Done\n')
                return 0
        
        if ignoreException == False:
            raise IxNetRestApiException('checkTrafficState: Traffic state did not reach the expected state(s): {0}. It is at: {1}'.format(
                expectedState, currentTrafficState))
        else:
            return 1

    def getRawTrafficItemSrcIp(self, trafficItemName):
        """
        Description
            Get the Raw Traffic Item source IP address. Mainly to look up each Device Group
            IPv4 that has the source IP address to get the gateway IP address.

        Parameter
            trafficItemName: The Raw Traffic Item name
        """
        trafficItemObj = self.getTrafficItemObjByName(trafficItemName)
        queryData = {
                    "from": trafficItemObj+"/configElement/1",
                    "nodes": [{"node": "stack", "properties": ["*"], "where": [{"property": "stackTypeId", "regex": "ipv4"}]},
                              {"node": "field", "properties": ["*"], "where": [{"property": "fieldTypeId", "regex": "ipv4.header.srcIp"}]}]
                    }
        queryResponse = self.ixnObj.query(data=queryData)
        sourceIp = queryResponse.json()['result'][0]['stack'][1]['field'][26]['fieldValue']
        return sourceIp

    def getTrafficItemType(self, trafficItemName):
        """
        Description
            Get the Traffic Item traffic type by the Traffic Item name.

        Parameter
            trafficItemName: The Traffic Item name

        Return
            The traffic type
        """
        queryData = {'from': '/traffic',
                    'nodes': [{'node': 'trafficItem', 'properties': ['name', 'trafficType'], 'where': [{'property': 'name', 'regex': trafficItemName}]},
                    ]}
        queryResponse = self.ixnObj.query(data=queryData)
        if queryResponse.json()['result'][0]['trafficItem'] == []:
            raise IxNetRestApiException('getTrafficItemType: No such Traffic Item Name found: %s' % trafficItemName)
        return (queryResponse.json()['result'][0]['trafficItem'][0]['trafficType'])

    def enableTrafficItemByName(self, trafficItemName, enable=True):
        """
        Description
            Enable or Disable a Traffic Item by its name.

        Parameter
            trafficItemName: The exact spelling of the Traffic Item name.
            enable: True | False
                    True: Enable Traffic Item
                    False: Disable Traffic Item
        """
        trafficItemObj = self.getTrafficItemObjByName(trafficItemName)
        if trafficItemObj == 0:
            raise IxNetRestApiException('No such Traffic Item name: %s' % trafficItemName)
        self.ixnObj.patch(self.ixnObj.httpHeader+trafficItemObj, data={"enabled": enable})

    def getTrafficItemName(self, trafficItemObj):
        """
        Description
            Get the Traffic Item name by its object.

        Parameter
            trafficItemObj: The Traffic Item object.
                            /api/v1/sessions/1/ixnetwork/traffic/trafficItem/1
        """
        response = self.ixnObj.get(self.ixnObj.httpHeader+trafficItemObj)
        return response.json()['name']

    def getAllTrafficItemObjects(self, getEnabledTrafficItemsOnly=False):
        """
        Description
            Get all the Traffic Item objects. 
            Due to the 100 items limit on REST call, using skip and take to break down the large list.
        
        Parameter
            getEnabledTrafficItemOnly: <bool>
        Return
            A list of Traffic Items
        """
        trafficItemObjList = []
        numOfTrafficItem = 0
        response = self.ixnObj.get(self.ixnObj.sessionUrl + '/traffic/trafficItem'  + "?skip=" + str(numOfTrafficItem) + "&take=100")

        while numOfTrafficItem < response.json()['count']:
            for eachTrafficItem in response.json()['data']:
                if getEnabledTrafficItemsOnly == True:
                    if eachTrafficItem['enabled'] == True:
                        trafficItemObjList.append(eachTrafficItem['links'][0]['href'])
                else:
                    trafficItemObjList.append(eachTrafficItem['links'][0]['href'])
            numOfTrafficItem += 100
            response = self.ixnObj.get(self.ixnObj.sessionUrl + '/traffic/trafficItem'  + "?skip=" + str(numOfTrafficItem) + "&take=100")
        
        return trafficItemObjList

    def getAllTrafficItemNames(self):
        """
        Description
            Return all of the Traffic Item names.
        """
        trafficItemUrl = self.ixnObj.sessionUrl+'/traffic/trafficItem'
        response = self.ixnObj.get(trafficItemUrl)
        trafficItemNameList = []
        for eachTrafficItemId in response.json():
            trafficItemNameList.append(eachTrafficItemId['name'])
        return trafficItemNameList

    def getTrafficItemObjByName(self, trafficItemName):
        """
        Description
            Get the Traffic Item object by the Traffic Item name.

        Parameter
            trafficItemName: Name of the Traffic Item.

        Return
            0: No Traffic Item name found. Return 0.
            traffic item object:  /api/v1/sessions/1/ixnetwork/traffic/trafficItem/2
        """
        queryData = {'from': '/traffic',
                    'nodes': [{'node': 'trafficItem', 'properties': ['name'], 'where': [{"property": "name", "regex": trafficItemName}]}
                    ]}
        queryResponse = self.ixnObj.query(data=queryData, silentMode=False)
        try:
            return queryResponse.json()['result'][0]['trafficItem'][0]['href']
        except:
            return 0

    def applyTraffic(self):
        """
        Description
            Apply the configured traffic.
        """
        restApiHeader = '/api'+self.ixnObj.sessionUrl.split('/api')[1]
        response = self.ixnObj.post(self.ixnObj.sessionUrl+'/traffic/operations/apply', data={'arg1': restApiHeader+'/traffic'})
        self.ixnObj.waitForComplete(response, self.ixnObj.sessionUrl+'/traffic/operations/apply/'+response.json()['id'])

    def regenerateTrafficItems(self, trafficItemList='all'):
        """
        Description
            Performs regenerate on Traffic Items.

        Parameter
            trafficItemList: 'all' will automatically regenerate from all Traffic Items.
                             Or provide a list of Traffic Items.
                             ['/api/v1/sessions/1/ixnetwork/traffic/trafficItem/1', ...]
        """
        if trafficItemList == 'all':
            response = self.ixnObj.get(self.ixnObj.sessionUrl + "/traffic/trafficItem")
            trafficItemList = ["%s" % (str(i["links"][0]["href"])) for i in response.json()]
        else:
            if type(trafficItemList) != list:
                trafficItemList = trafficItemList.split(' ')

        url = self.ixnObj.sessionUrl+"/traffic/trafficItem/operations/generate"
        data = {"arg1": trafficItemList}
        self.ixnObj.logInfo('Regenerating traffic items: %s' % trafficItemList)
        response = self.ixnObj.post(url, data=data)
        self.ixnObj.waitForComplete(response, url+'/'+response.json()['id'])

    def startTraffic(self, regenerateTraffic=True, applyTraffic=True, blocking=False):
        """
        Description
            Start traffic and verify traffic is started.
            This function will also give you the option to regenerate and apply traffic.

        Parameter
            regenerateTraffic: <bool>
                          
            applyTraffic: <bool> 
                          In a situation like packet capturing, you cannot apply traffic after
                          starting packet capture because this will stop packet capturing. 
                          You need to set applyTraffic to False in this case.

            blocking: <bool> If True, API server doesn't return until it has
                             started traffic and ready for stats.  Unblocking is the opposite.

        Syntax
            For blocking state:
               POST:  /api/v1/sessions/{id}/ixnetwork/traffic/operations/startstatelesstrafficblocking'
               DATA:  {arg1: ['/api/v1/sessions/{id}/ixnetwork/traffic/trafficItem/{id}' ...]}

            For non blocking state:
               POST: /api/v1/sessions/1/ixnetwork/traffic/operations/start
               DATA: {arg1: '/api/v1/sessions/{id}/ixnetwork/traffic'}

        Requirements:
            For non blocking state only:

               # You need to check the traffic state before getting stats.
               # Note: Use the configElementObj returned by configTrafficItem()
               if trafficObj.getTransmissionType(configElementObj) == "fixedFrameCount":
                   trafficObj.checkTrafficState(expectedState=['stopped', 'stoppedWaitingForStats'], timeout=45)

               if trafficObj.getTransmissionType(configElementObj) == "continuous":
                   trafficObj.checkTrafficState(expectedState=['started', 'startedWaitingForStats'], timeout=45)
        """
        if regenerateTraffic:
            self.regenerateTrafficItems()

        if applyTraffic:
            self.applyTraffic()

        if blocking == False:
            url = self.ixnObj.sessionUrl+'/traffic/operations/start'
            response = self.ixnObj.post(url, data={'arg1': self.ixnObj.sessionUrl+'/traffic'})
            self.ixnObj.waitForComplete(response, url + '/' + response.json()['id'], timeout=120)

        # Server will go into blocking state until it is ready to accept the next api command.
        if blocking == True:
            enabledTrafficItemList = self.getAllTrafficItemObjects(getEnabledTrafficItemsOnly=True)
            url = self.ixnObj.sessionUrl+'/traffic/trafficItem/operations/startstatelesstrafficblocking'
            response = self.ixnObj.post(url, data={'arg1': enabledTrafficItemList})
            self.ixnObj.waitForComplete(response, url + '/' + response.json()['id'], timeout=120)

    def stopTraffic(self, blocking=False):
        """
        Description
            Stop traffic and verify traffic has stopped.

        Parameters
           blocking: <bool>: True=Synchronous mode. Server will not accept APIs until the process is complete.

        Syntax
            For blocking state:
               POST: /api/v1/sessions/{id}/ixnetwork/traffic/operations/stopstatelesstrafficblocking
               DATA:  {arg1: ['/api/v1/sessions/{id}/ixnetwork/traffic/trafficItem/{id}' ...]}

            For non blocking state:
               POST: /api/v1/sessions/{id}/ixnetwork/traffic/operations/stop
               DATA: {'arg1': '/api/v1/sessions/{id}/ixnetwork/traffic'}
        """
        if blocking == True:
            #queryData = {"from": "/traffic",
            #    "nodes": [{"node": "trafficItem", "properties": ["enabled"], "where": [{"property": "enabled", "regex": "True"}]}]}

            #queryResponse = self.ixnObj.query(data=queryData, silentMode=False)
            #enabledTrafficItemHrefList = [trafficItem['href'] for trafficItem in queryResponse.json()['result'][0]['trafficItem']]

            enabledTrafficItemList = self.getAllTrafficItemObjects(getEnabledTrafficItemsOnly=True)
            url = self.ixnObj.sessionUrl+'/traffic/operations/stopstatelesstrafficblocking'
            response = self.ixnObj.post(url, data={'arg1': enabledTrafficItemList})
            self.ixnObj.waitForComplete(response, url + '/' + response.json()['id'], timeout=120)

        if blocking == False:
            self.ixnObj.logInfo('stopTraffic: %s' % self.ixnObj.sessionUrl+'/traffic/operations/stop')
            url = self.ixnObj.sessionUrl+'/traffic/operations/stop'            
            response = self.ixnObj.post(url, data={'arg1': '{0}/ixnetwork/traffic'.format(self.ixnObj.headlessSessionId)})
            self.ixnObj.waitForComplete(response, url + '/' + response.json()['id'], timeout=120)

        self.checkTrafficState(expectedState=['stopped'])
        time.sleep(3)

    def showTrafficItems(self):
        """
        Description
            Show All Traffic Item details.
        """
        queryData = {'from': '/traffic',
                    'nodes': [{'node': 'trafficItem',    'properties': ['name', 'enabled', 'state', 'biDirectional', 'trafficType', 'warning', 'errors'], 'where': []},
                                {'node': 'endpointSet',   'properties': ['name', 'sources', 'destinations'], 'where': []},
                                {'node': 'configElement', 'properties': ['name', 'endpointSetId', ], 'where': []},
                                {'node': 'frameSize',     'properties': ['type', 'fixedSize'], 'where': []},
                                {'node': 'framePayload',  'properties': ['type', 'customRepeat'], 'where': []},
                                {'node': 'frameRate',     'properties': ['type', 'rate'], 'where': []},
                                {'node': 'frameRateDistribution', 'properties': ['streamDistribution', 'portDistribution'], 'where': []},
                                {'node': 'transmissionControl', 'properties': ['type', 'frameCount', 'burstPacketCount'], 'where': []},
                                {'node': 'tracking',      'properties': ['trackBy'], 'where': []},
                            ]
                    }

        queryResponse = self.ixnObj.query(data=queryData, silentMode=True)
        self.ixnObj.logInfo('\n', end='', timestamp=False)
        for ti in queryResponse.json()['result'][0]['trafficItem']:
            self.ixnObj.logInfo('TrafficItem: {0}\n\tName: {1}  Enabled: {2}  State: {3}'.format(
                ti['id'], ti['name'], ti['enabled'], ti['state']), timestamp=False)
            self.ixnObj.logInfo('\tTrafficType: {0}  BiDirectional: {1}'.format(ti['trafficType'],
                                                                                ti['biDirectional']),
                                timestamp=False)

            for tracking in ti['tracking']:
                self.ixnObj.logInfo('\tTrackings: {0}'.format(tracking['trackBy']), timestamp=False)

            for endpointSet, cElement in zip(ti['endpointSet'], ti['configElement']):
                self.ixnObj.logInfo('\tEndpointSetId: {0}  EndpointSetName: {1}'.format(endpointSet['id'],
                                                                                        endpointSet['name']), timestamp=False)
                srcList = []
                for src in endpointSet['sources']:
                    srcList.append(src.split('/ixnetwork')[1])

                dstList = []
                for dest in endpointSet['destinations']:
                    dstList.append(dest.split('/ixnetwork')[1])

                self.ixnObj.logInfo('\t    Sources: {0}'.format(srcList), timestamp=False)
                self.ixnObj.logInfo('\t    Destinations: {0}'.format(dstList), timestamp=False)
                self.ixnObj.logInfo('\t    FrameType: {0}  FrameSize: {1}'.format(cElement['frameSize']['type'],
                                                                                  cElement['frameSize']['fixedSize']),
                                    timestamp=False)
                self.ixnObj.logInfo('\t    TranmissionType: {0}  FrameCount: {1}  BurstPacketCount: {2}'.format(
                    cElement['transmissionControl']['type'],
                    cElement['transmissionControl']['frameCount'],
                    cElement['transmissionControl']['burstPacketCount']),
                                    timestamp=False)

                self.ixnObj.logInfo('\t    FrameRateType: {0}  FrameRate: {1}'.format(
                    cElement['frameRate']['type'], cElement['frameRate']['rate']), timestamp=False)

            self.ixnObj.logInfo('\n', end='', timestamp=False)

    def setFrameSize(self, trafficItemName, **kwargs):
        """
        Description
            Modify the frame size.

        Parameters
            type: <str>:  fixed|increment|presetDistribution|quadGaussian|random|weightedPairs
        
            trafficItemName: <str>: The name of the Traffic Item..

        Example:
            trafficObj.setFrameSize('Topo1 to Topo2', type='fxied', fixedSize=128)
            trafficObj.setFrameSize('Topo1 to Topo2', type='increment', incrementFrom=68, incrementStep=2, incrementTo=1200)
        """
        queryData = {'from': '/traffic',
                    'nodes': [{'node': 'trafficItem', 'properties': ['name'], 'where': [{'property': 'name', 'regex': trafficItemName}]},
                              {'node': 'configElement', 'properties': [], 'where': []}]}
        queryResponse = self.ixnObj.query(data=queryData)
        if queryResponse.json()['result'][0]['trafficItem'] == []:
            raise IxNetRestApiException('\nNo such Traffic Item name found: %s' % trafficItemName)

        configElementObj = queryResponse.json()['result'][0]['trafficItem'][0]['configElement'][0]['href']
        self.ixnObj.patch(self.ixnObj.httpHeader+configElementObj+'/frameSize', data=kwargs)

    def configFramePayload(self, configElementObj, payloadType='custom', customRepeat=True, customPattern=None):
        """
        Description
            Configure the frame payload.

        Parameters
            payloadType: <str>: Options:
                           custom, decrementByte, decrementWord, incrementByte, incrementWord, random
            customRepeat: <bool>
            customPattern: <str>: Enter a custom payload pattern
        """
        data = {'type': payloadType, 'customRepeat': customRepeat, 'customPattern': customPattern}
        self.ixnObj.patch(self.ixnObj.httpHeader+configElementObj+'/framePayload', data=data)

    def enableMinFrameSize(self, enable=True):
        """
        Description
           Enable the global traffic option to allow smaller frame size.

        Parameter
           enable: <bool>: True to enable it.
        """
        self.ixnObj.patch(self.ixnObj.sessionUrl+'/traffic', data={'enableMinFrameSize': enable})

    def suspendTrafficItem(self, trafficItemObj, suspend=True):
        """
        Description
           Suspend the Traffic Item from sending traffic.
        
        Parameter
           trafficItemObj: <str>: The Traffic Item object.
           suspend: <bool>: True=suspend traffic.
        
        Syntax
           PATCH: /api/v1/sessions/{id}/ixnetwork/traffic/trafficItem/{id}
           DATA:  {'suspend': True}
        """
        self.ixnObj.patch(self.ixnObj.httpHeader+trafficItemObj, data={'suspend': suspend})






"""
Description
     Demonstrate how to use Python class inheritance to leverage OpenIxia
     library and extend the OpenIxia library with your own library.

     Create your own classes and inherit the corresponding OpenIxia class.
     Your scripts will be instantiating objects using your classes.
     With inheritance, this allows your scripts to use all of OpenIxia functions
     and use your added functions in your classes.
     This allows you to keep getting OpenIxia updates without interferring with 
     your extended functions.

"""

from IxNetRestApi import *
from IxNetRestApiPortMgmt import PortMgmt
from IxNetRestApiTraffic import Traffic
from IxNetRestApiProtocol import Protocol
from IxNetRestApiStatistics import Statistics
from IxNetRestApiPacketCapture import PacketCapture
from IxNetRestApiQuickTest import QuickTest

class Connection(Connect):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

class Port_Mgmt(PortMgmt):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def my_Own_Port_Mgmt_Function(self):
        # Add and create your own function to extend the OpenIxia library.
        pass

    def exportJsonConfig(self):
        # Overwrite the OpenIxia exportJsonConfig function.
        # For enhancement or to fix a bug in exportJsonConfig().
        pass

class Traffic_Config(Traffic):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

class Protocol_Config(Protocol):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

class Statistics_View(Statistics):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



import re, time
from .IxNetRestApi import IxNetRestApiException
from .IxNetRestApiFileMgmt import FileMgmt

class Statistics(object):
    def __init__(self, ixnObj=None):
        self.ixnObj = ixnObj

        # For takesnapshot()
        self.fileMgmtObj = FileMgmt(self.ixnObj)

    def setMainObject(self, mainObject):
        """
        Description
            For Python Robot Framework support.
        """
        self.ixnObj = mainObject

    def getStats(self, viewObject=None, viewName='Flow Statistics', csvFile=None, csvEnableFileTimestamp=False,
                 displayStats=True, silentMode=True, ignoreError=False):
        """
        Description
           Get stats for any viewName.
           The method calls two different methods based on the IxNetwork version that you are using.
           For IxNetwork version prior to 8.50, calls getStatsPage.
           For IxNetwork version >= 8.50, calls getStatsData. This has new APIs that is more robust and they don't
           work in versions prior to 8.50.

        Parameters
            csvFile = None or <filename.csv>.
                      None will not create a CSV file.
                      Provide a <filename>.csv to record all stats to a CSV file.
                      Example: getStats(sessionUrl, csvFile='Flow_Statistics.csv')

            csvEnableFileTimestamp = True or False. If True, timestamp will be appended to the filename.
            displayStats: True or False. True=Display stats.
            ignoreError: True or False.  Returns None if viewName is not found.
            viewObject: The view object: http://{apiServerIp:port}/api/v1/sessions/2/ixnetwork/statistics/view/13
                        A view object handle could be obtained by calling getViewObject().

            viewName options (Not case sensitive):
               NOTE: Not all statistics are listed here.
                  You could get the statistic viewName directly from the IxNetwork GUI in the statistics.
        """
        buildNumber = float(self.ixnObj.getIxNetworkVersion()[:3])
        if buildNumber >= 8.5:
            return self.getStatsData(viewObject=viewObject, viewName=viewName, csvFile=csvFile, csvEnableFileTimestamp=csvEnableFileTimestamp,
                                     displayStats=displayStats, silentMode=silentMode, ignoreError=ignoreError)
        else:
            return self.getStatsPage(viewObject=viewObject, viewName=viewName, csvFile=csvFile, csvEnableFileTimestamp=csvEnableFileTimestamp,
                                     displayStats=displayStats, silentMode=silentMode, ignoreError=ignoreError)
            
    def getStatsPage(self, viewObject=None, viewName='Flow Statistics', csvFile=None, csvEnableFileTimestamp=False,
                     displayStats=True, silentMode=True, ignoreError=False):
        """
        Description
            Get stats by the statistic name or get stats by providing a view object handle.
            This method uses deprecated APIs effective IxNetwork version 8.50: /statistics/statView/<id>/page
            Starting 8.50, the new API to use is: /statistics/statView/<id>/data, which is in getStatsData()

        Parameters
            csvFile = None or <filename.csv>.
                      None will not create a CSV file.
                      Provide a <filename>.csv to record all stats to a CSV file.
                      Example: getStats(sessionUrl, csvFile='Flow_Statistics.csv')

            csvEnableFileTimestamp = True or False. If True, timestamp will be appended to the filename.

            displayStats: True or False. True=Display stats.

            ignoreError: True or False.  Returns None if viewName is not found.

            viewObject: The view object: http://{apiServerIp:port}/api/v1/sessions/2/ixnetwork/statistics/view/13
                        A view object handle could be obtained by calling getViewObject().

            viewName options (Not case sensitive):
               NOTE: Not all statistics are listed here.
                  You could get the statistic viewName directly from the IxNetwork GUI in the statistics.

            'Port Statistics'
            'Tx-Rx Frame Rate Statistics'
            'Port CPU Statistics'
            'Global Protocol Statistics'
            'Protocols Summary'
            'Port Summary'
            'BGP Peer Per Port'
            'OSPFv2-RTR Drill Down'
            'OSPFv2-RTR Per Port'
            'IPv4 Drill Down'
            'L2-L3 Test Summary Statistics'
            'Flow Statistics'
            'Traffic Item Statistics'
            'IGMP Host Drill Down'
            'IGMP Host Per Port'
            'IPv6 Drill Down'
            'MLD Host Drill Down'
            'MLD Host Per Port'
            'PIMv6 IF Drill Down'
            'PIMv6 IF Per Port'
            'Flow View'

         Note: Not all of the viewNames are listed here. You have to get the exact names from
               the IxNetwork GUI in statistics based on your protocol(s).

         Return a dictionary of all the stats: statDict[rowNumber][columnName] == statValue
           Get stats on row 2 for 'Tx Frames' = statDict[2]['Tx Frames']
        """
        if viewObject == None:
            breakFlag = 0
            counterStop = 30
            for counter in range(1, counterStop+1):
                viewList = self.ixnObj.get('%s/%s/%s' % (self.ixnObj.sessionUrl, 'statistics', 'view'), silentMode=silentMode)
                views = ['%s/%s/%s/%s' % (self.ixnObj.sessionUrl, 'statistics', 'view', str(i['id'])) for i in viewList.json()]
                if silentMode is False:
                    self.ixnObj.logInfo('\ngetStats: Searching for viewObj for viewName: %s' % viewName)

                for view in views:
                    # GetAttribute
                    response = self.ixnObj.get('%s' % view, silentMode=silentMode)
                    captionMatch = re.match(viewName, response.json()['caption'], re.I)
                    if captionMatch:
                        # viewObj: sessionUrl + /statistics/view/11'
                        viewObject = view
                        breakFlag = 1
                        break

                if breakFlag == 1:
                    break

                if counter < counterStop:
                    self.ixnObj.logInfo('\nGetting statview [{0}] is not ready. Waiting {1}/{2} seconds.'.format(
                        viewName, counter, counterStop), timestamp=False)
                    time.sleep(1)
                    continue
                    
                if counter == counterStop:
                    if viewObject == None and ignoreError == False:
                        raise IxNetRestApiException("viewObj wasn't found for viewName: {0}".format(viewName))

                    if viewObject == None and ignoreError == True:
                        return None

        if silentMode is False:
            self.ixnObj.logInfo('\n[{0}] viewObj is: {1}'.format(viewName, viewObject))

        for counter in range(0,31):
            response = self.ixnObj.get(viewObject+'/page', silentMode=silentMode)
            totalPages = response.json()['totalPages']
            if totalPages == 'null':
                self.ixnObj.logInfo('\nGetting total pages is not ready yet. Waiting %d/30 seconds' % counter)
                time.sleep(1)

            if totalPages != 'null':
                break

            if counter == 30:
                raise IxNetRestApiException('getStatsPage failed: Getting total pages')

        if csvFile != None:
            import csv
            csvFileName = csvFile.replace(' ', '_')
            if csvEnableFileTimestamp:
                import datetime
                timestamp = datetime.datetime.now().strftime('%H%M%S')
                if '.' in csvFileName:
                    csvFileNameTemp = csvFileName.split('.')[0]
                    csvFileNameExtension = csvFileName.split('.')[1]
                    csvFileName = csvFileNameTemp+'_'+timestamp+'.'+csvFileNameExtension
                else:
                    csvFileName = csvFileName+'_'+timestamp

            csvFile = open(csvFileName, 'w')
            csvWriteObj = csv.writer(csvFile)

        # Get the stat column names
        columnList = response.json()['columnCaptions']
        if csvFile != None:
            csvWriteObj.writerow(columnList)

        flowNumber = 1
        statDict = {}
        # Get the stat values
        for pageNumber in range(1,totalPages+1):
            self.ixnObj.patch(viewObject+'/page', data={'currentPage': pageNumber}, silentMode=silentMode)

            response = self.ixnObj.get(viewObject+'/page', silentMode=silentMode)
            statValueList = response.json()['pageValues']

            for statValue in statValueList:
                if csvFile != None:
                    csvWriteObj.writerow(statValue[0])

                if displayStats:
                    self.ixnObj.logInfo('\nRow: %d' % flowNumber, timestamp=False)

                statDict[flowNumber] = {}
                index = 0
                for statValue in statValue[0]:
                    statName = columnList[index]
                    statDict[flowNumber].update({statName: statValue})
                    if displayStats:
                        self.ixnObj.logInfo('\t%s: %s' % (statName, statValue), timestamp=False)
                    index += 1
                flowNumber += 1

        if csvFile != None:
            csvFile.close()
        return statDict

    def getStatsData(self, viewObject=None, viewName='Flow Statistics', csvFile=None, csvEnableFileTimestamp=False,
                     displayStats=True, silentMode=False, ignoreError=False):
        """
        Description
            For IxNetwork version >= 8.50.   
            Get stats by the statistic name or get stats by providing a view object handle.
            This method get stats using /api/v1/sessions/{id}/ixnetwork/statistics/view/{id}/data to get
            attributes columnCaptions, pageValues and totalPages. This method uses new API starting in 
            version 8.50.

        Parameters
            csvFile = None or <filename.csv>.
                      None will not create a CSV file.
                      Provide a <filename>.csv to record all stats to a CSV file.
                      Example: getStats(sessionUrl, csvFile='Flow_Statistics.csv')

            csvEnableFileTimestamp = True or False. If True, timestamp will be appended to the filename.
            displayStats: True or False. True=Display stats.
            ignoreError: True or False.  Returns None if viewName is not found.
            viewObject: The view object: http://{apiServerIp:port}/api/v1/sessions/2/ixnetwork/statistics/view/13
                        A view object handle could be obtained by calling getViewObject().

            viewName options (Not case sensitive):
               NOTE: Not all statistics are listed here.
                  You could get the statistic viewName directly from the IxNetwork GUI in the statistics.

            'Port Statistics'
            'Tx-Rx Frame Rate Statistics'
            'Port CPU Statistics'
            'Global Protocol Statistics'
            'Protocols Summary'
            'Port Summary'
            'BGP Peer Per Port'
            'OSPFv2-RTR Drill Down'
            'OSPFv2-RTR Per Port'
            'IPv4 Drill Down'
            'L2-L3 Test Summary Statistics'
            'Flow Statistics'
            'Traffic Item Statistics'
            'IGMP Host Drill Down'
            'IGMP Host Per Port'
            'IPv6 Drill Down'
            'MLD Host Drill Down'
            'MLD Host Per Port'
            'PIMv6 IF Drill Down'
            'PIMv6 IF Per Port'
            'Flow View'

         Note: Not all of the viewNames are listed here. You have to get the exact names from
               the IxNetwork GUI in statistics based on your protocol(s).

         Return a dictionary of all the stats: statDict[rowNumber][columnName] == statValue
           Get stats on row 2 for 'Tx Frames' = statDict[2]['Tx Frames']
        """
        if viewObject == None:

            breakFlag = 0
            counterStop = 30
            for counter in range(1, counterStop+1):
                viewList = self.ixnObj.get('%s/%s/%s' % (self.ixnObj.sessionUrl, 'statistics', 'view'), silentMode=silentMode)
                views = ['%s/%s/%s/%s' % (self.ixnObj.sessionUrl, 'statistics', 'view', str(i['id'])) for i in viewList.json()]
                if silentMode is False:
                    self.ixnObj.logInfo('\ngetStats: Searching for viewObj for viewName: {0}'.format(
                        viewName), timestamp=False)

                for view in views:
                    #print('\nview:', view)
                    response = self.ixnObj.get('%s' % view, silentMode=True)
                    captionMatch = re.match(viewName, response.json()['caption'], re.I)
                    if captionMatch:
                        # viewObj: sessionUrl + /statistics/view/11'
                        viewObject = view
                        breakFlag = 1
                        break

                if breakFlag == 1:
                    break

                if counter < counterStop:
                    self.ixnObj.logInfo('\nGetting statview [{0}] is not ready. Waiting {1}/{2} seconds.'.format(
                        viewName, counter, counterStop), timestamp=False)
                    time.sleep(1)
                    continue
                    
                if counter == counterStop:
                    if viewObject == None and ignoreError == False:
                        raise IxNetRestApiException("viewObj wasn't found for viewName: {0}".format(viewName))

                    if viewObject == None and ignoreError == True:
                        return None

        if silentMode is False:
            self.ixnObj.logInfo('\n[{0}] viewObj is: {1}'.format(viewName, viewObject))

        counterStop = 30
        for counter in range(0, counterStop+1):
            response = self.ixnObj.get(viewObject+'/data', silentMode=silentMode)
            totalPages = response.json()['totalPages']
            #self.ixnObj.logInfo('totalPages: {0}'.format(totalPages), timestamp=False)

            if totalPages == 'null' and counter < counterStop:
                self.ixnObj.logInfo('\nGetting total pages is not ready yet. Waiting {0}/{1} seconds'.format(
                    counter, counterStop), timestamp=False)
                time.sleep(1)
                continue

            if totalPages != 'null' and counter < counterStop:
                break

            if counter == counterStop:
                raise IxNetRestApiException('getStats failed: Getting total pages')

        if csvFile != None:
            import csv
            csvFileName = csvFile.replace(' ', '_')
            if csvEnableFileTimestamp:
                import datetime
                timestamp = datetime.datetime.now().strftime('%H%M%S')
                if '.' in csvFileName:
                    csvFileNameTemp = csvFileName.split('.')[0]
                    csvFileNameExtension = csvFileName.split('.')[1]
                    csvFileName = csvFileNameTemp+'_'+timestamp+'.'+csvFileNameExtension
                else:
                    csvFileName = csvFileName+'_'+timestamp

            csvFile = open(csvFileName, 'w')
            csvWriteObj = csv.writer(csvFile)

        flowNumber = 1
        statDict = {}
        getColumnCaptionFlag = 0
        for pageNumber in range(1,totalPages+1):
            self.ixnObj.patch(viewObject+'/data', data={'currentPage': pageNumber}, silentMode=silentMode)

            counterStop = 30
            for counter in range(1, counterStop+1):
                response = self.ixnObj.get(viewObject+'/data', silentMode=silentMode)
                if counter < counterStop:
                    if response.json()['columnCaptions'] == [] or response.json()['pageValues'] == []:
                        self.ixnObj.logInfo('[{0}] stat values not ready yet. Wait {1}/{2} seconds.'.format(
                            viewName, counter, counterStop))
                        time.sleep(1)
                        continue

                    if response.json()['columnCaptions'] != [] or response.json()['pageValues'] != []:
                        break

                if counter == counterStop:
                    raise IxNetRestApiException('IxNetwork API server failed to provide stats')

            # Get the stat column names one time only
            if getColumnCaptionFlag == 0:
                getColumnCaptionFlag = 1
                columnList = response.json()['columnCaptions']
                if csvFile != None:
                    csvWriteObj.writerow(columnList)
                
            statValueList = response.json()['pageValues']
            for statValue in statValueList:
                if csvFile != None:
                    csvWriteObj.writerow(statValue[0])

                if displayStats:
                    self.ixnObj.logInfo('\nRow: %d' % flowNumber, timestamp=False)

                statDict[flowNumber] = {}
                index = 0
                for statValue in statValue[0]:
                    statName = columnList[index]
                    statDict[flowNumber].update({statName: statValue})
                    if displayStats:
                        self.ixnObj.logInfo('\t%s: %s' % (statName, statValue), timestamp=False)
                    index += 1
                flowNumber += 1

        if csvFile != None:
            csvFile.close()
        return statDict

    def removeAllTclViews(self):
        """
        Description
           Removes all created stat views.
        """
        removeAllTclViewsUrl = self.ixnObj.sessionUrl+'/operations/removealltclviews'
        response = self.ixnObj.post(removeAllTclViewsUrl)
        self.ixnObj.waitForComplete(response, removeAllTclViewsUrl+'/'+response.json()['id'])

    def takeSnapshot(self, viewName='Flow Statistics', windowsPath=None, isLinux=False, localLinuxPath=None,
                     renameDestinationFile=None, includeTimestamp=False, mode='overwrite'):
        """
        Description
            Take a snapshot of the vieweName statistics.  This is a two step process.
            1> Take a snapshot of the statistics that you want and store it in the C: drive for Windows.
               For Linux, the snapshot goes to /home/ixia_logs.
            2> Copy the statistics from the snapshot locations to the local Linux where you ran the script..

        Parameters
            viewName: The name of the statistics to get.
            windowsPath: For Windows|WindowsConnectionMgr only.
                         The C: drive + path to store the snapshot: Example: c:\\Results.
            isLinux: <bool>: Defaults to False.  Set to True if you're getting the snapshot from Linux chassis.
            localLinuxPath: None|path. Provide the local Linux path to put the snapshot file.
                            If None, this API won't copy the stat file to local Linux.
                            The stat file will remain on Windows c: drive.
            renameDestinationFile: None or a name of the file other than the viewName.
            includeTimestamp: True|False: To include a timestamp at the end of the file.
            mode: append|overwrite: append=To append stats to an existing stat file.
                                    overwrite=Don't append stats. Create a new stat file.

        Example:
            For Windows:
               statObj.takeSnapshot(viewName='Flow Statistics', windowsPath='C:\\Results', localLinuxPath='/home/hgee',
                           renameDestinationFile='my_renamed_stat_file.csv', includeTimestamp=True)

            For Linux:
               statObj.takeSnapshot(viewName='Flow Statistics', isLinux=True, localLinuxPath='/home/hgee')
        """
        if mode == 'append':
            mode = 'kAppendCSVFile'

        if mode == 'overwrite':
            mode = 'kOverwriteCSVFile'

        if windowsPath:
            location = windowsPath

        if isLinux:
            location = '/home/ixia_logs'

        data = {'arg1': [viewName], 'arg2': [
                            "Snapshot.View.Contents: \"allPages\"",
                            "Snapshot.View.Csv.Location: \"{0}\"".format(location),
                            "Snapshot.View.Csv.GeneratingMode: \"%s\"" % mode,
                            "Snapshot.View.Csv.StringQuotes: \"True\"",
                            "Snapshot.View.Csv.SupportsCSVSorting: \"False\"",
                            "Snapshot.View.Csv.FormatTimestamp: \"True\"",
                            "Snapshot.View.Csv.DumpTxPortLabelMap: \"False\"",
                            "Snapshot.View.Csv.DecimalPrecision: \"3\""
                            ]
                }

        url = self.ixnObj.sessionUrl+'/operations/takeviewcsvsnapshot'
        response = self.ixnObj.post(url, data=data)
        self.ixnObj.waitForComplete(response, url+'/'+response.json()['id'])
        if isLinux:
            snapshotFile = location + '/' + viewName + '.csv'
            self.fileMgmtObj.copyFileLinuxToLocalLinux(linuxApiServerPathAndFileName=snapshotFile, localPath=localLinuxPath,
                                                        renameDestinationFile=renameDestinationFile,
                                                        includeTimestamp=includeTimestamp)

        if windowsPath and localLinuxPath:
            # Get the snapshot. Use the csvFilename that was specified and the location
            self.fileMgmtObj.copyFileWindowsToLocalLinux('{0}\\{1}.csv'.format(windowsPath, viewName), localLinuxPath,
                                                         renameDestinationFile=renameDestinationFile,
                                                         includeTimestamp=includeTimestamp)

    def getViewObject(self, viewName='Flow Statistics'):
        """
        Description
            To get just the statistic view object.
            Mainly used by internal APIs such as takeCsvSnapshot that
            requires the statistics view object handle.

        Parameter
         viewName:  Options (case sensitive):
            "Port Statistics"
            "Tx-Rx Frame Rate Statistics"
            "Port CPU Statistics"
            "Global Protocol Statistics"
            "Protocols Summary"
            "Port Summary"
            "OSPFv2-RTR Drill Down"
            "OSPFv2-RTR Per Port"
            "IPv4 Drill Down"
            "L2-L3 Test Summary Statistics"
            "Flow Statistics"
            "Traffic Item Statistics"
        """
        self.ixnObj.logInfo('\ngetStats: %s' % viewName)
        viewList = self.ixnObj.get("%s/%s/%s" % (self.ixnObj.sessionUrl, "statistics", "view"))
        views = ["%s/%s/%s/%s" % (self.ixnObj.sessionUrl, "statistics", "view", str(i["id"])) for i in viewList.json()]
        for view in views:
            # GetAttribute
            response = self.ixnObj.get(view)
            caption = response.json()["caption"]
            if viewName == caption:
                # viewObj: sessionUrl + "/statistics/view/11"
                viewObj = view
                return viewObj
        return None

    def clearStats(self):
        """
        Description
            Clear all stats and wait for API server to finish.
        """
        url = self.ixnObj.sessionUrl + '/operations/clearStats'
        response = self.ixnObj.post(url)



# PLEASE READ DISCLAIMER
#
#    This class demonstrates sample IxNetwork REST API usage for
#    demo and reference purpose only.
#    It is subject to change for updates without warning.
#
# Description
#    A class object for IxNetwork Classic Framework.

from .IxNetRestApi import IxNetRestApiException
import requests, json, os, re, sys, time, datetime, ast

class ClassicProtocol(object):
    def __init__(self, ixnObj=None):
        """
        Parameters
           ixnObj: <str>: The main connection object.
        """
        self.ixnObj = ixnObj

        from IxNetRestApiPortMgmt import PortMgmt
        self.portMgmtObj = PortMgmt(self.ixnObj)

        from IxNetRestApiStatistics import Statistics
        self.statObj = Statistics(self.ixnObj)

    def getPortsByProtocol(self, protocolName):
        """
        Description
            Based on the specified protocol, return all ports associated with the protocol.

        Parameters
           protocolName options:
              bfd, bgp, cfm, eigrp, elmi, igmp, isis, lacp, ldp, linkOam, lisp, mld,
              mplsOam, mplsTp, openFlow, ospf, ospfV3, pimsm, ping, rip, ripng, rsvp,
              static, stp

        Returns: [chassisIp, cardNumber, portNumber]
                  Example: [['192.168.70.11', '1', '1'], ['192.168.70.11', '1', '2']]

        Returns [] if no port is configured with the specified protocolName
        """
        portList = []
        response = self.ixnObj.get(self.ixnObj.sessionUrl+'/vport')
        # ['http://{apiServerIp:port}/api/v1/sessions/1/ixnetwork/vport/1']
        vportList = ['%s/%s/%s' % (self.ixnObj.sessionUrl, 'vport', str(i["id"])) for i in response.json()]

        # Go through each port that has the protocol enabled.
        for vport in vportList:
            # http://{apiServerIp:port}/api/v1/sessions/1/ixnetwork/vport/1/protocols/ospf
            currentProtocol = vport+'/protocols/'+protocolName
            response = self.ixnObj.get(currentProtocol)
            if response.json()['enabled'] == True:
                # 192.168.70.11:1:5
                response = self.ixnObj.get(vport)
                assignedTo = response.json()['assignedTo']
                currentChassisIp  = str(assignedTo.split(':')[0])
                currentCardNumber = str(assignedTo.split(':')[1])
                currentPortNumber = str(assignedTo.split(':')[2])
                currentPort = [currentChassisIp, currentCardNumber, currentPortNumber]
                portList.append(currentPort)

        return portList

    def getProtocolListByPort(self, port):
        """
        Description
            Get all enabled protocols by the specified port.

        Parameters
            port: [chassisIp, cardNumber, portNumber] -> ['192.168.70.11', '2', '8']
        """
        self.ixnObj.logInfo('\ngetProtocolListByPort...')
        chassis = str(port[0])
        card = str(port[1])
        port = str(port[2])
        specifiedPort = [chassis, card, port]
        enabledProtocolList = []
        response = self.ixnObj.get(self.ixnObj.sessionUrl+'/vport')
        vportList = ['%s/%s/%s' % (self.ixnObj.sessionUrl, 'vport', str(i["id"])) for i in response.json()]
        for vport in vportList:
            response = self.ixnObj.get(vport, 'assignedTo')
            # 192.168.70.11:1:5
            assignedTo = response.json()['assignedTo']
            currentChassisIp  = str(assignedTo.split(':')[0])
            currentCardNumber = str(assignedTo.split(':')[1])
            currentPortNumber = str(assignedTo.split(':')[2])
            currentPort = [currentChassisIp, currentCardNumber, currentPortNumber]
            if currentPort != specifiedPort:
                continue
            else:
                response = self.ixnObj.get(vport+'/protocols?links=true')
                if response.status_code == 200:
                     #print 'json', response.json()['links']
                    for protocol in response.json()['links']:
                        currentProtocol = protocol['href']
                        url = self.ixnObj.httpHeader+currentProtocol
                        response = self.ixnObj.get(url)
                        if 'enabled' in response.json() and response.json()['enabled'] == True:
                            # Exclude ARP object
                            if 'arp' not in currentProtocol:
                                enabledProtocolList.append(str(currentProtocol))

        return enabledProtocolList

    def sendArpOnPort(self, portName):
        """
        Description
            Send Arp request on a specified port

        Parameters
            portName: <str>: Name of the port. eg: '1/1/11'

        Syntax
            POST: http://10.154.162.94:11009/api/v1/sessions/1/ixnetwork/vport/operations/sendarp
            DATA: {"arg1": "/api/v1/sessions/1/ixnetwork/vport/1"}

        Examples
            sendArpOnPort(portName='1/1/11')
        """
        vport = self.portMgmtObj.getVportObjectByName(portName)
        if vport == None:
            raise IxNetRestApiException('PortName {0} not connected to chassis'.format(portName))

        url = self.ixnObj.sessionUrl + '/vport/operations/sendarp'
        response = self.ixnObj.post(url, data={'arg1': vport})
        self.ixnObj.waitForComplete(response, url + '/' + response.json()['id'])

    def getDiscoverdNeighborOnPort(self, portName):
        """
        Description
            Get Arp discovered neighbor on a specified port

        Parameters
            portName: <str>: Name of the port. eg: '1/1/11'

        Syntax
            GET: http://10.154.162.94:11009/api/v1/sessions/1/ixnetwork/vport/1/discoveredNeighbor/1

        Examples
            getDiscoverdNeighborOnPort(portName='1/1/11')

        Return
            Discovered mac address
        """
        vport = self.portMgmtObj.getVportObjectByName(portName)
        if vport == None:
            raise IxNetRestApiException('PortName {0} not connected to chassis'.format(portName))

        url = vport + '/discoveredNeighbor'
        response = self.ixnObj.get(url)

        url = self.ixnObj.httpHeader + response.json()[0]['links'][0]['href']
        response = self.ixnObj.get(url)

        self.ixnObj.logInfo('Discovered Neighbor: %s' % response.json()['neighborMac'])
        return response.json()['neighborMac']


    def startStopProtocolOnPort(self, protocol, portName, action='start'):
        """
        Description
            Start and stop a protocol on a specified port

        Parameters
            protocol: <str>: Protocol to start
            portName: <str>: Name of the port, eg: '1/1/11'
            action: <str>: start or stop a protocol, default is start

        Syntax
            POST: /api/v1/sessions/{id}/ixnetwork/vport/protocols/<protocol>
            DATA: {'args': 'api/v1/sessions/1/ixnetwork/vport/1'}

        Examples
            startStopProtocolOnPort(protocol='ospf', portName='1/1/11')
            startStopProtocolOnPort(protocol='ospf', portName='1/1/11', action='stop')

        """
        vport = self.portMgmtObj.getVportObjectByName(portName)
        if vport == None:
            raise IxNetRestApiException('PortName {0} not connected to chassis'.format(portName))

        url = self.ixnObj.sessionUrl + '/vport/protocols/' + protocol + '/operations/' + action
        vport = vport + '/protocols/' + protocol
        response = self.ixnObj.post(url, data={'arg1': vport})
        self.ixnObj.waitForComplete(response, url + '/' + response.json()['id'])

    def getConfiguredProtocols(self):
        """
        Description
            Get the list of protocols configured on vports.

        Return
            A list or one or more congfigured protocols eg: "['ospf','bgp']"
            return [] if no protocol is configured
        """
        time.sleep(5)
        configuredProtocolList = []
        protocolList = ['bfd', 'bgp', 'eigrp', 'isis', 'ldp', 'lisp',
                        'mplsOam', 'mplsTp', 'ospf', 'ospfV3', 'pimsm',
                        'rip', 'ripng']
        response = self.ixnObj.get(self.ixnObj.sessionUrl + '/vport')
        if response == False:
            raise IxNetRestApiException('No ports connected to chassis')

        vportList = ['%s' % vport['links'][0]['href'] for vport in response.json()]
        for eachVport in vportList:
            for eachProtocol in protocolList:
                node = '/router'
                if re.search('bgp', eachProtocol, re.I):
                    node = '/neighborRange'
                protocolResponse = self.ixnObj.get(self.ixnObj.httpHeader + eachVport + '/protocols/' + eachProtocol)
                response = self.ixnObj.get(self.ixnObj.httpHeader + eachVport + '/protocols/' + eachProtocol + node)
                if response.json() != []:
                    if response.json()[0]['enabled'] == True and protocolResponse.json()['runningState'] == 'started':
                        configuredProtocolList.append(eachProtocol)
        configuredProtocolList = list(set(configuredProtocolList))
        return configuredProtocolList

    def enableProtocolOnPort(self, protocol, portName, enable=True):
        """
        Description
           Enable protocol on a speficied port

        Parameters
            protocol: <str>: Protocol to start
            portName: <str>: Name of the port eg: '1/1/11'
            enable: <bool> enable or disable specified protocol. default is True

        Syntax
            PATCH: /api/v1/sessions/{id}/ixnetwork/vport/{id}/protocols/<protocol>/router/{id}
            PATCH: /api/v1/sessions/{id}/ixnetwork/vport/{id}/protocols/<protocol>/neighborRange/{id}
            PATCH: /api/v1/sessions/{id}/ixnetwork/vport/{id}/protocols/<protocol>/host/{id}
            PATCH: /api/v1/sessions/{id}/ixnetwork/vport/{id}/protocols/<protocol>/bridge/{id}
            PATCH: /api/v1/sessions/{id}/ixnetwork/vport/{id}/protocols/<protocol>/actor/{id}
            PATCH: /api/v1/sessions/{id}/ixnetwork/vport/{id}/protocols/<protocol>/neighborPair/{id}

        Examples
            enableProtocolOnPort(protocol='ospf', portName='1/1/11')
            enableProtocolOnPort(protocol='ospf', portName='1/1/11', enable=False)
        """
        vport = self.portMgmtObj.getVportObjectByName(portName)
        if vport == None:
            raise IxNetRestApiException('PortName {0} not connected to chassis'.format(portName))

        RouterInstanceList = self.getRouterInstanceByPortAndProtocol(protocol=protocol, vport=vport)
        if RouterInstanceList == []:
            raise IxNetRestApiException('No Router instance exists in protocol {0}'.format(protocol))

        for eachRouterInstance in RouterInstanceList:
            url = self.ixnObj.httpHeader + eachRouterInstance
            self.ixnObj.patch(url, data={"enabled": enable})

    def getProtocolSessionsStats(self, portName, protocol):
        """
        Description
            Get a protocol session status for a specified protocol on a port

        Parameters
            portName: <str>: Name of the Port to get the protocol session stats eg: "1/1/11"
            protocol: <str>: Name of the protocol. eg: <ospf/ospfV3/bgp/isis/ripng/bfd/rip/ldp/mplsoam/pim>

        Examples
            getProtocolSessionsStats(portName='1/1/11', protocol='ospf')

        Return
            Protocol stats in dictionary format eg: {'ospf': {'Configured': 2, 'Up': 1}}
        """
        protocolStats = {}
        vport = self.portMgmtObj.getVportObjectByName(portName)
        if vport == None:
            raise IxNetRestApiException('PortName {0} not connected to chassis'.format(portName))

        node = '/router'
        if re.search('bgp', protocol, re.I):
            node = '/neighborRange'
        protocolResponse = self.ixnObj.get(vport + '/protocols/' + protocol)
        response = self.ixnObj.get(vport + '/protocols/' + protocol + node)

        if response.json() != []:
            if response.json()[0]['enabled'] == True and protocolResponse.json()['runningState'] == 'started':
                if re.search('ospf', protocol, re.I):
                    protocolViewName = 'OSPF Aggregated Statistics'
                elif re.search('ospfV3', protocol, re.I):
                    protocolViewName = 'OSPFv3 Aggregated Statistics'
                elif re.search('bgp', protocol, re.I):
                    protocolViewName = 'BGP Aggregated Statistics'
                elif re.search('isis', protocol, re.I):
                    protocolViewName = 'ISIS Aggregated Statistics'
                elif re.search('ripng', protocol, re.I):
                    protocolViewName = 'RIPng Aggregated Statistics'
                elif re.search('bfd', protocol, re.I):
                    protocolViewName = 'BFD Aggregated Statistics'
                elif re.search('rip', protocol, re.I):
                    protocolViewName = 'RIP Aggregated Statistics'
                elif re.search('ldp', protocol, re.I):
                    protocolViewName = 'LDP Aggregated Statistics'
                elif re.search('mplsoam', protocol, re.I):
                    protocolViewName = 'MPLSOAM Aggregated Statistics'
                elif re.search('pim', protocol, re.I):
                    protocolViewName = 'PIMSM Aggregated Statistics'
                else:
                    raise IxNetRestApiException('No viewName defined')
            else:
                raise IxNetRestApiException('No {0} protocol running or enabled on port {1}'.format(protocol, portName))
        else:
            raise IxNetRestApiException('No {0} protocol configured on port {1}'.format(protocol, portName))

        stats = self.statObj.getStats(viewName=protocolViewName, displayStats=False)

        #totalPorts = len(stats.keys());  # Length stats.keys() represents total ports.
        self.ixnObj.logInfo('ProtocolViewName: {0}'.format(protocolViewName))
        for session in stats.keys():
            if re.search('OSPF', protocolViewName, re.I):
                sessionsUp = int(stats[session]['Full Nbrs.'])
                totalSessions = int(stats[session]['Sess. Configured'])
            elif re.search('BGP', protocolViewName, re.I):
                sessionsUp = int(stats[session]['Sess. Up'])
                totalSessions = int(stats[session]['Sess. Configured'])
            elif re.search('ISIS', protocolViewName, re.I):
                sessionsUp = int(stats[session]['L2 Sess. Up'])
                totalSessions = int(stats[session]['L2 Sess. Configured'])
            elif re.search('RIPng', protocolViewName, re.I) or re.search('BFD', protocolViewName, re.I):
                sessionsUp = int(stats[session]['Routers Running'])
                totalSessions = int(stats[session]['Routers Configured'])
            elif re.search('RIP', protocolViewName, re.I):
                sessionsUp = int(stats[session]['Request Packet Tx'])
                totalSessions = int(stats[session]['Routers Configured'])
            elif re.search('LACP', protocolViewName, re.I):
                sessionsUp = int(stats[session]['LAG Member Ports UP'])
                totalSessions = int(stats[session]['Total LAG Member Ports'])
            elif re.search('LDP', protocolViewName, re.I):
                sessionsUp = int(stats[session]['Targeted Sess. Up'])
                totalSessions = int(stats[session]['Targeted Sess. Configured'])
            elif re.search('MPLS', protocolViewName, re.I):
                sessionsUp = int(stats[session]['BFD Up-Sessions'])
                totalSessions = int(stats[session]['BFD Session Count'])
            elif re.search('PIM', protocolViewName, re.I):
                sessionsUp = int(stats[session]['Rtrs. Running'])
                totalSessions = int(stats[session]['Rtrs. Configured'])
                # totalSessionsNotStarted = int(stats[session]['Sessions Not Started'])
            else:
                raise IxNetRestApiException('No protocol viewName found')

            if stats[session]['Port Name'] == portName:
                self.ixnObj.logInfo('\n\tPortName: {0}\n\t   TotalSessionsUp: {1}\n\t   TotalSessionsConfigured: {2}'.format(
                    stats[session]['Port Name'], sessionsUp, totalSessions), timestamp=False)
                protocolStats[protocol] = {'Configured': totalSessions, 'Up': sessionsUp}
        return protocolStats

    def enableRouteRangeOnProtocol(self, portName, protocol, routeRange, enable=True):
        """
        Description
            Enable a route range for a protocol on a speficied port

        Parameters
            portName: <str>: Name of the port eg: "1/1/11"
            protocol: <str>: protocol to enable route range. eg: <OSPF|OSPFV3|BGP|ISIS|EIGRP|RIP|RIPng>
            routeRange: <str>: route range <IPv4|IPv6> address
            enable: <bool>: enable or disable route range, default is True (enable)

        Syntax
            PATCH: /api/v1/sessions/{id}/ixnetwork/vport/{id}/protocols/<protocol>/router/{id}/routeRange/{id}
            PATCH: /api/v1/sessions/{id}/ixnetwork/vport/{id}/protocols/<protocol>/neighborRange/{id}/routeRange/{id}

        Examples:
            enableRouteRangeOnProtocol(protName='1/1/11', protocol='ospf', routeRange='10.10.10.1')
            enableRouteRangeOnProtocol(protName='1/1/11', protocol='ospfv3', routeRange='10::1', enable=True)
        """
        vport = self.portMgmtObj.getVportObjectByName(portName)
        if vport == None:
            raise IxNetRestApiException('PortName {0} not connected to chassis'.format(portName))

        RouterInstanceList = self.getRouterInstanceByPortAndProtocol(protocol=protocol, vport=vport)
        if RouterInstanceList == []:
            raise IxNetRestApiException('No Router instance exists in protocol {0}'.format(protocol))

        args = 'firstRoute'
        if protocol == 'ospf':
            args = 'networkNumber'
        if protocol == 'bgp':
            args = 'networkAddress'

        for eachRouterInstance in RouterInstanceList:
            url = self.ixnObj.httpHeader + eachRouterInstance + '/routeRange'
            response = self.ixnObj.get(url)
            RouteRangeInstanceList = ["%s" % (str(i["links"][0]["href"])) for i in response.json()]
            self.ixnObj.logInfo('Route Range list %s' % RouteRangeInstanceList)

            for eachRouteRange in RouteRangeInstanceList:
                url = self.ixnObj.httpHeader + eachRouteRange
                response = self.ixnObj.get(url)
                RouteRangeNetwork = response.json()[args]
                if RouteRangeNetwork == routeRange:
                    self.ixnObj.patch(url, data={"enabled": enable})
                    return
            raise IxNetRestApiException(
                'Route range: {0} does not exist in protocol: {1} port: {2}'.format(routeRange, protocol, portName))

    def removeRouteRangeOnProtocol(self, portName, protocol, routeRange):
        """
        Description
            Remove a route range for a protocol on a speficied port

        Parameters
            portName: <str>: Name of the port eg: "1/1/11"
            protocol: <str>: protocol to remove route range. eg: <OSPF|OSPFV3|BGP|ISIS|EIGRP|RIP|RIPng>
            routeRange: <str>: route range <IPv4|IPv6> address

        Syntax
            DELETE: /api/v1/sessions/{id}/ixnetwork/vport/{id}/protocols/<protocol>/router/{id}/routeRange/{id}
            DELETE: /api/v1/sessions/{id}/ixnetwork/vport/{id}/protocols/<protocol>/neighborRange/{id}/routeRange/{id}

        Examples:
            removeRouteRangeOnProtocol(protName='1/1/11', protocol='ospf', routeRange='10.10.10.1')
            removeRouteRangeOnProtocol(protName='1/1/11', protocol='ospfv3', routeRange='10::1')
        """
        vport = self.portMgmtObj.getVportObjectByName(portName)
        if vport == None:
            raise IxNetRestApiException('PortName {0} not connected to chassis'.format(portName))

        RouterInstanceList = self.getRouterInstanceByPortAndProtocol(protocol=protocol, vport=vport)
        if RouterInstanceList == []:
            raise IxNetRestApiException('No Router instance exists in protocol {0}'.format(protocol))

        self.ixnObj.logInfo('Router list %s' % RouterInstanceList)
        args = 'firstRoute'
        if protocol == 'ospf':
            args = 'networkNumber'
        if protocol == 'bgp':
            args = 'networkAddress'

        for eachRouterInstance in RouterInstanceList:
            url = self.ixnObj.httpHeader + eachRouterInstance + '/routeRange'
            response = self.ixnObj.get(url)
            RouteRangeInstanceList = ["%s" % (str(i["links"][0]["href"])) for i in response.json()]
            self.ixnObj.logInfo('Route Range list %s' % RouteRangeInstanceList)
            for eachRouteRange in RouteRangeInstanceList:
                url = self.ixnObj.httpHeader + eachRouteRange
                response = self.ixnObj.get(url)
                RouteRangeNetwork = response.json()[args]
                if RouteRangeNetwork == routeRange:
                    self.ixnObj.delete(url)
                    return

            raise IxNetRestApiException(
                'Route range: {0} does not exist in protocol: {1} port: {2}'.format(routeRange, protocol, portName))

    def createRouteRangeOnProtocol(self, portName, protocol, routeRange):
        """
        Description
            Create a route range for a protocol on a speficied port

        Parameters
            portName: <str>: Name of the port eg: "1/1/11"
            protocol: <str>: protocol to create route range. eg: <OSPF|OSPFV3|BGP|ISIS|EIGRP|RIP|RIPng>
            routeRange: <dict>: route range to configure <IPv4|IPv6> address
            eg: {'enabled': 'True', 'mask': 24, 'numberOfRoutes': 5, 'networkNumber': '8.7.7.1', 'metric': 10, 'origin': 'externalType1'}
                {'enabled': 'True', 'maskWidth': 64, 'numberOfRoute': 10, 'firstRoute': '7::1', 'metric': 10, 'nextHop': '7::2'}

        Syntax
            POST: /api/v1/sessions/{id}/ixnetwork/vport/{id}/protocols/ospf/router/{id}/routeRange/{id}
            DATA: {'enabled': 'True', 'mask': 24, 'numberOfRoutes': 5, 'networkNumber': '7.7.7.1', 'metric': 10, 'origin': 'externalType1'}

            POST: /api/v1/sessions/{id}/ixnetwork/vport/{id}/protocols/ospfV3/router/{id}/routeRange/{id}
            DATA: {'enabled': 'True', 'mask': 64, 'numberOfRoutes': 5, 'firstRoute': '7::1', 'metric': 10, 'type': 'anotherArea', 'addressFamily': 'unicast'}

            POST: /api/v1/sessions/{id}/ixnetwork/vport/{id}/protocols/eigrp/router/{id}/routeRange/{id}
            DATA: {'enabled': 'True', 'mask': 24, 'numberOfRoutes': 10, 'firstRoute': '7.7.7.1', 'metric': 10, 'nextHop': '7.7.7.2'}

            POST: /api/v1/sessions/{id}/ixnetwork/vport/{id}/protocols/rip/router/{id}/routeRange/{id}
            DATA: {'enabled': 'True', 'maskWidth': 24, 'noOfRoutes': 10, 'firstRoute': '7.7.7.1', 'metric': 10, 'nextHop': '7.7.7.2'}

            POST: /api/v1/sessions/{id}/ixnetwork/vport/{id}/protocols/ripng/router/{id}/routeRange/{id}
            DATA: {'enabled': 'True', 'maskWidth': 64, 'numberOfRoute': 10, 'firstRoute': '7::1', 'metric': 10, 'nextHop': '7::2'}

            POST: /api/v1/sessions/{id}/ixnetwork/vport/{id}/protocols/bgp/neighborRange/{id}/routeRange/{id}
            DATA: {'enabled': 'True', 'fromPrefix': 24, 'thruPrefix': 24, 'numRoutes': 10, 'networkAddress': '7.7.7.7'}

            POST: /api/v1/sessions/{id}/ixnetwork/vport/{id}/protocols/bgp/neighborRange/{id}/routeRange/{id}
            DATA: {'enabled': 'True', 'fromPrefix': 64, 'thruPrefix': 64, 'numRoutes': 10, 'networkAddress': '7::1'}

        Examples
            createRouteRangeOnProtocol(portName='1/1/11', protocol='ospf', routeRange={'enabled': 'True', 'mask': 24,
                                                                         'numberOfRoutes': 5, 'networkNumber': '8.7.7.1',
                                                                         'metric': 10, 'origin': 'externalType1'}
            createRouteRangeOnProtocol(portName='1/1/11', protocol='ospf', routeRange={'networkNumber': '8.7.7.1'}

        """
        vport = self.portMgmtObj.getVportObjectByName(portName)
        if vport == None:
            raise IxNetRestApiException('PortName {0} not connected to chassis'.format(portName))

        RouterInstanceList = self.getRouterInstanceByPortAndProtocol(protocol=protocol, vport=vport)
        if RouterInstanceList == []:
            raise IxNetRestApiException('No Router instance exists in protocol {0}'.format(protocol))

        self.ixnObj.logInfo('Router list %s' % RouterInstanceList)
        #routeRange = ast.literal_eval(routeRange)
        for eachRouterInstance in RouterInstanceList:
            url = self.ixnObj.httpHeader + eachRouterInstance + '/routeRange'
            self.ixnObj.post(url, data=routeRange)

    def getRouterInstanceByPortAndProtocol(self, protocol, vport):
        """
        Description
            Get router instance for the specified protocol on a speficied vport

        Parameters
            protocol: <str>: protocol to get a router instance
            vport: <str>: vport instance eg: "/api/v1/sessions/1/ixnetwork/vport/1"

        Syntax
            GET: /api/v1/sessions/{id}/ixnetwork/vport/{id}/protocols/<protocol>/router
            GET: /api/v1/sessions/{id}/ixnetwork/vport/{id}/protocols/<protocol>/neighborRange
            GET: /api/v1/sessions/{id}/ixnetwork/vport/{id}/protocols/<protocol>/host
            GET: /api/v1/sessions/{id}/ixnetwork/vport/{id}/protocols/<protocol>/bridge
            GET: /api/v1/sessions/{id}/ixnetwork/vport/{id}/protocols/<protocol>/actor
            GET: /api/v1/sessions/{id}/ixnetwork/vport/{id}/protocols/<protocol>/neighborPair

        Examples
            getRouterInstanceByPortAndProtocol(protocol='ospf', vport='/api/v1/sessions/1/ixnetwork/vport/1')

        Return
            RouterInstanceList
            Returns [] if no router instance exists
        """
        if protocol == 'bgp':
            nextNode = '/neighborRange'
        elif protocol == 'igmp' or protocol == 'mld':
            nextNode = '/host'
        elif protocol == 'stp':
            nextNode = '/bridge'
        elif protocol == 'rsvp':
            nextNode = '/neighborPair'
        elif protocol == 'lacp':
            nextNode = '/link'
        else:
            nextNode = '/router'
        url = vport+'/protocols/'+protocol+nextNode
        response = self.ixnObj.get(url)
        RouterInstanceList = ["%s" % (str(i["links"][0]["href"])) for i in response.json()]
        self.ixnObj.logInfo('Router Instance list %s' % RouterInstanceList)
        return RouterInstanceList

    def verifyProtocolSessionsUp(self, protocolViewName='BGP Peer Per Port', timeout=60):
        """
        Description
            Verify the specified protocol sessions are UP or not.

        Parameters
            protocolViewName: <str>: The protocol view name. Get this name from API browser or in IxNetwork GUI statistic tabs.
            timeout: <int>: Duration to wait for the protocol sessions to up. Default = 60 seconds.

        protocolViewName options:
            'BFD Aggregated Statistics'
            'BGP Aggregated Statistics'
            'ISIS Aggregated Statistics'
            'OSPF Aggregated Statistics'
            'RIPng Aggregated Statistics'
            'RIP Aggregated Statistics'
            'LDP Aggregated Statistics'
            'PIMSM Aggregated Statistics'
            'MPLSOAM Aggregated Statistics'

        Examples
            verifyProtocolSessionsUp(protcolViewName='ospf Aggregated Statistics')
            verifyProtocolSessionsUp(protcolViewName='ospf Aggregated Statistics',timeout=90)
        """
        totalSessionsDetectedUp = 0
        totalSessionsDetectedDown = 0
        totalPortsUpFlag = 0
        self.ixnObj.logInfo('Protocol view name %s' % protocolViewName)
        time.sleep(10)
        for counter in range(1, timeout + 1):
            stats = self.statObj.getStats(viewName=protocolViewName, displayStats=False)
            totalPorts = len(stats.keys())  # Length stats.keys() represents total ports.
            self.ixnObj.logInfo('ProtocolName: {0}'.format(protocolViewName))
            for session in stats.keys():
                if re.search('OSPF', protocolViewName, re.I):
                    sessionsUp = int(stats[session]['Full Nbrs.'])
                    totalSessions = int(stats[session]['Sess. Configured'])
                elif re.search('BGP', protocolViewName, re.I):
                    sessionsUp = int(stats[session]['Sess. Up'])
                    totalSessions = int(stats[session]['Sess. Configured'])
                elif re.search('ISIS', protocolViewName, re.I):
                    sessionsUp = int(stats[session]['L2 Sess. Up'])
                    totalSessions = int(stats[session]['L2 Sess. Configured'])
                elif re.search('RIPng', protocolViewName, re.I) or re.search('BFD', protocolViewName, re.I):
                    sessionsUp = int(stats[session]['Routers Running'])
                    totalSessions = int(stats[session]['Routers Configured'])
                elif re.search('RIP', protocolViewName, re.I):
                    sessionsUp = int(stats[session]['Request Packet Tx'])
                    totalSessions = int(stats[session]['Routers Configured'])
                elif re.search('LACP', protocolViewName, re.I):
                    sessionsUp = int(stats[session]['LAG Member Ports UP'])
                    totalSessions = int(stats[session]['Total LAG Member Ports'])
                elif re.search('LDP', protocolViewName, re.I):
                    sessionsUp = int(stats[session]['Targeted Sess. Up'])
                    totalSessions = int(stats[session]['Targeted Sess. Configured'])
                elif re.search('MPLS', protocolViewName, re.I):
                    sessionsUp = int(stats[session]['BFD Up-Sessions'])
                    totalSessions = int(stats[session]['BFD Session Count'])
                elif re.search('PIM', protocolViewName, re.I):
                    sessionsUp = int(stats[session]['Rtrs. Running'])
                    totalSessions = int(stats[session]['Rtrs. Configured'])
                # totalSessionsNotStarted = int(stats[session]['Sessions Not Started'])
                totalExpectedSessionsUp = totalSessions

                if totalExpectedSessionsUp != 0:
                    self.ixnObj.logInfo(
                        '\n\tPortName: {0}\n\t   TotalSessionsUp: {1}\n\t   ExpectedTotalSessionsup: {2}'.format(
                            stats[session]['Port Name'], sessionsUp, totalExpectedSessionsUp), timestamp=False)
                    if counter < timeout and sessionsUp != totalExpectedSessionsUp:
                        self.ixnObj.logInfo('\t   Protocol Session is still down', timestamp=False)

                    if counter < timeout and sessionsUp == totalExpectedSessionsUp:
                        totalPortsUpFlag += 1
                        if totalPortsUpFlag == totalPorts:
                            self.ixnObj.logInfo('All protocol sessions are up!')
                            return

            if counter == timeout and sessionsUp != totalExpectedSessionsUp:
                raise IxNetRestApiException('Protocol Sessions failed to come up')

            self.ixnObj.logInfo('\n\tWait {0}/{1} seconds\n'.format(counter, timeout), timestamp=False)
            time.sleep(1)

    def verifyAllConfiguredProtocolSessions(self, duration):
        """
        Description
            verify all configured protocol sessions are UP or not

        Parameters
            duration: <int>: duration to wait for the protocol sessions to UP

        Examples
            verifyAllConfiguredProtocolSessions(duration=120)
            verifyAllConfiguredProtocolSessions(120)
        """
        response = self.getConfiguredProtocols()
        if response == []:
            raise IxNetRestApiException('No protocols Running or Configured or Enabled')

        for eachProtocol in response:
            if re.search('ospf', eachProtocol,re.I):
                viewName = 'OSPF Aggregated Statistics'
            elif re.search('ospfV3', eachProtocol,re.I):
                viewName = 'OSPFv3 Aggregated Statistics'
            elif re.search('bgp', eachProtocol,re.I):
                viewName = 'BGP Aggregated Statistics'
            elif re.search('isis', eachProtocol,re.I):
                viewName = 'ISIS Aggregated Statistics'
            elif re.search('ripng', eachProtocol,re.I):
                viewName = 'RIPng Aggregated Statistics'
            elif re.search('bfd', eachProtocol,re.I):
                viewName = 'BFD Aggregated Statistics'
            elif re.search('rip', eachProtocol,re.I):
                viewName = 'RIP Aggregated Statistics'
            elif re.search('ldp', eachProtocol,re.I):
                viewName = 'LDP Aggregated Statistics'
            elif re.search('mplsoam', eachProtocol,re.I):
                viewName = 'MPLSOAM Aggregated Statistics'
            elif re.search('pim', eachProtocol,re.I):
                viewName = 'PIMSM Aggregated Statistics'
            else:
                raise IxNetRestApiException('No viewName defined')

            self.verifyProtocolSessionsUp(protocolViewName=viewName, timeout=duration)



import re, time
from .IxNetRestApi import IxNetRestApiException
from .IxNetRestApiFileMgmt import FileMgmt

class QuickTest(object):
    def __init__(self, ixnObj=None, fileMgmtObj=None):
        self.ixnObj = ixnObj
        if fileMgmtObj:
            self.fileMgmtObj = fileMgmtObj
        else:
            self.fileMgmtObj = FileMgmt(ixnObj)

    def setMainObject(self, mainObject):
        # For Python Robot Framework support
        self.ixnObj = mainObject
        self.fileMgmtObj.setMainObject(mainObject)

    def getAllQuickTestHandles(self):
        """
        Description
            Get all the Quick Test object handles

        Returns:
            ['/api/v1/sessions/1/ixnetwork/quickTest/rfc2544throughput/2',
             '/api/v1/sessions/1/ixnetwork/quickTest/rfc2889broadcastRate/1',
             '/api/v1/sessions/1/ixnetwork/quickTest/rfc2889broadcastRate/2']
        """
        response = self.ixnObj.get(self.ixnObj.sessionUrl+'/quickTest')
        quickTestHandles = []
        for eachTestId in response.json()['testIds']:
            quickTestHandles.append(eachTestId)
        return quickTestHandles

    def getAllQuickTestNames(self):
        quickTestNameList = []
        for eachQtHandle in self.getAllQuickTestHandles():
            response = self.ixnObj.get(self.ixnObj.httpHeader+eachQtHandle)
            quickTestNameList.append(response.json()['name'])
        return quickTestNameList

    def getQuickTestHandleByName(self, quickTestName):
        """
        Description
            Get the Quick Test object handle by the name.

        Parameter
            quickTestName: The name of the Quick Test.
        """
        for quickTestHandle in self.getAllQuickTestHandles():
            response = self.ixnObj.get(self.ixnObj.httpHeader+quickTestHandle)
            currentQtName = response.json()['name']
            if (bool(re.match(quickTestName, currentQtName, re.I))):
                return quickTestHandle

    def getQuickTestNameByHandle(self, quickTestHandle):
        """
        quickTestHandle = /api/v1/sessions/1/ixnetwork/quickTest/rfc2544throughput/2
        """
        response = self.ixnObj.get(self.ixnObj.httpHeader + quickTestHandle)
        return response.json()['name']

    def getQuickTestDuration(self, quickTestHandle):
        """
        quickTestHandle = /api/v1/sessions/1/ixnetwork/quickTest/rfc2544throughput/2
        """
        response = self.ixnObj.get(self.ixnObj.httpHeader + quickTestHandle + '/testConfig')
        return response.json()['duration']

    def getQuickTestTotalFrameSizesToTest(self, quickTestHandle):
        """
        quickTestHandle = /api/v1/sessions/1/ixnetwork/quickTest/rfc2544throughput/2
        """
        response = self.ixnObj.get(self.ixnObj.httpHeader + quickTestHandle + '/testConfig')
        return response.json()['framesizeList']

    def applyQuickTest(self, qtHandle):
        """
        Description
            Apply Quick Test configurations

        Parameter
            qtHandle: The Quick Test object handle
        """
        response = self.ixnObj.post(self.ixnObj.sessionUrl+'/quickTest/operations/apply', data={'arg1': qtHandle})
        if self.ixnObj.waitForComplete(response, self.ixnObj.sessionUrl+'/quickTest/operations/apply/'+response.json()['id']) == 1:
            raise IxNetRestApiException('applyTraffic: waitForComplete failed')

    def getQuickTestCurrentAction(self, quickTestHandle):
        """
        quickTestHandle = /api/v1/sessions/1/ixnetwork/quickTest/rfc2544throughput/2
        """
        ixNetworkVersion = self.ixnObj.getIxNetworkVersion()
        match = re.match('([0-9]+)\.[^ ]+ *', ixNetworkVersion)
        if int(match.group(1)) >= 8:
            timer = 10
            for counter in range(1,timer+1):
                response = self.ixnObj.get(self.ixnObj.httpHeader+quickTestHandle+'/results', silentMode=True)
                if counter < timer and response.json()['currentActions'] == []:
                    self.ixnObj.logInfo('getQuickTestCurrentAction is empty. Waiting %s/%s' % (counter, timer))
                    time.sleep(1)
                    continue
                if counter < timer and response.json()['currentActions'] != []:
                    break
                if counter == timer and response.json()['currentActions'] == []:
                    IxNetRestApiException('getQuickTestCurrentActions: Has no action')

            return response.json()['currentActions'][-1]['arg2']
        else:
            response = self.ixnObj.get(self.ixnObj.httpHeader+quickTestHandle+'/results')
            return response.json()['progress']

    def verifyQuickTestInitialization(self, quickTestHandle):
        """
        quickTestHandle = /api/v1/sessions/1/ixnetwork/quickTest/rfc2544throughput/2
        """
        for timer in range(1,20+1):
            currentAction = self.getQuickTestCurrentAction(quickTestHandle)
            print('verifyQuickTestInitialization currentState: %s' % currentAction)
            if timer < 20:
                if currentAction == 'TestEnded' or currentAction == 'None':
                    self.ixnObj.logInfo('\nverifyQuickTestInitialization CurrentState = %s\n\tWaiting %s/20 seconds to change state' % (currentAction, timer))
                    time.sleep(1)
                    continue
                else:
                    break
            if timer >= 20:
                if currentAction == 'TestEnded' or currentAction == 'None':
                    self.ixnObj.showErrorMessage()
                    raise IxNetRestApiException('Quick Test is stuck at TestEnded.')

        ixNetworkVersionNumber = int(self.ixnObj.getIxNetworkVersion().split('.')[0])
        applyQuickTestCounter = 60
        for counter in range(1,applyQuickTestCounter+1):
            quickTestApplyStates = ['InitializingTest', 'ApplyFlowGroups', 'SetupStatisticsCollection']
            currentAction = self.getQuickTestCurrentAction(quickTestHandle)
            if currentAction == None:
                currentAction = 'ApplyingAndInitializing'

            print('\nverifyQuickTestInitialization: %s  Expecting: TransmittingFrames\n\tWaiting %s/%s seconds' % (currentAction, counter, applyQuickTestCounter))
            if ixNetworkVersionNumber >= 8:
                if counter < applyQuickTestCounter and currentAction != 'TransmittingFrames':
                    time.sleep(1)
                    continue

                if counter < applyQuickTestCounter and currentAction == 'TransmittingFrames':
                    self.ixnObj.logInfo('\nVerifyQuickTestInitialization is done applying configuration and has started transmitting frames\n')
                break

            if ixNetworkVersionNumber < 8:
                if counter < applyQuickTestCounter and currentAction == 'ApplyingAndInitializing':
                    time.sleep(1)
                    continue

                if counter < applyQuickTestCounter and currentAction == 'ApplyingAndInitializing':
                    self.ixnObj.logInfo('\nVerifyQuickTestInitialization is done applying configuration and has started transmitting frames\n')
                break

            if counter == applyQuickTestCounter:
                if ixNetworkVersionNumber >= 8 and currentAction != 'TransmittingFrames':
                    self.ixnObj.showErrorMessage()
                    if currentAction == 'ApplyFlowGroups':
                        self.ixnObj.logInfo('\nIxNetwork is stuck on Applying Flow Groups. You need to go to the session to FORCE QUIT it.\n')
                    raise IxNetRestApiException('\nVerifyQuickTestInitialization is stuck on %s. Waited %s/%s seconds' % (
                            currentAction, counter, applyQuickTestCounter))

                if ixNetworkVersionNumber < 8 and currentAction != 'Trial':
                    self.ixnObj.showErrorMessage()
                    raise IxNetRestApiException('\nVerifyQuickTestInitialization is stuck on %s. Waited %s/%s seconds' % (
                            currentAction, counter, applyQuickTestCounter))

    def startQuickTest(self, quickTestHandle):
        """
        Description
            Start a Quick Test

        Parameter
            quickTestHandle: The Quick Test object handle.
            /api/v1/sessions/{id}/ixnetwork/quickTest/rfc2544throughput/2

        Syntax
           POST: http://{apiServerIp:port}/api/v1/sessions/{1}/ixnetwork/quickTest/operations/start
                 data={arg1: '/api/v1/sessions/{id}/ixnetwork/quickTest/rfc2544throughput/2'}
                 headers={'content-type': 'application/json'}
        """
        url = self.ixnObj.sessionUrl+'/quickTest/operations/start'
        self.ixnObj.logInfo('\nstartQuickTest:%s' % url)
        response = self.ixnObj.post(url, data={'arg1': quickTestHandle})
        if self.ixnObj.waitForComplete(response, url+'/'+response.json()['id']) == 1:
            raise IxNetRestApiException

    def stopQuickTest(self, quickTestHandle):
        """
        Description
            Stop the Quick Test.

        Parameter
            quickTestHandle: The Quick Test object handle.
            /api/v1/sessions/{id}/ixnetwork/quickTest/rfc2544throughput/2

        Syntax
           POST: http://{apiServerIp:port}/api/v1/sessions/{1}/ixnetwork/quickTest/operations/stop
                 data={arg1: '/api/v1/sessions/{id}/ixnetwork/quickTest/rfc2544throughput/2'}
                 headers={'content-type': 'application/json'}
        """
        url = self.ixnObj.sessionUrl+'/quickTest/operations/stop'
        response = self.ixnObj.post(url, data={'arg1': quickTestHandle})
        if self.ixnObj.waitForComplete(response, url+'/'+response.json()['id']) == 1:
            raise IxNetRestApiException

    def monitorQuickTestRunningProgress(self, quickTestHandle, getProgressInterval=10):
        """
        Description
            monitor the Quick Test running progress.
            For Linux API server only, it must be a NGPF configuration. (Classic Framework is not supported in REST)

        Parameters
            quickTestHandle: /api/v1/sessions/{1}/ixnetwork/quickTest/rfc2544throughput/2
        """
        isRunningBreakFlag = 0
        trafficStartedFlag = 0
        waitForRunningProgressCounter = 0
        counter = 1

        while True:
            response = self.ixnObj.get(self.ixnObj.httpHeader+quickTestHandle+'/results', silentMode=True)
            isRunning = response.json()['isRunning']
            if isRunning == True:
                response = self.ixnObj.get(self.ixnObj.httpHeader+quickTestHandle+'/results', silentMode=True)
                currentRunningProgress = response.json()['progress']
                if bool(re.match('^Trial.*', currentRunningProgress)) == False:
                    if waitForRunningProgressCounter < 30:
                        self.ixnObj.logInfo('isRunning=True. Waiting for trial runs {0}/30 seconds'.format(waitForRunningProgressCounter))
                        waitForRunningProgressCounter += 1
                        time.sleep(1)
                    if waitForRunningProgressCounter == 30:
                        raise IxNetRestApiException('isRunning=True. No quick test stats showing.')
                else:
                    trafficStartedFlag = 1
                    self.ixnObj.logInfo(currentRunningProgress)
                    counter += 1
                    time.sleep(getProgressInterval)
                    continue
            else:
                if trafficStartedFlag == 1:
                    # We only care about traffic not running in the beginning.
                    # If traffic ran and stopped, then break out.
                    self.ixnObj.logInfo('\nisRunning=False. Quick Test is complete')
                    return 0
                if isRunningBreakFlag < 20:
                    print('isRunning=False. Wait {0}/20 seconds'.format(isRunningBreakFlag))
                    isRunningBreakFlag += 1
                    time.sleep(1)
                    continue
                if isRunningBreakFlag == 20:
                    raise IxNetRestApiException('Quick Test failed to start:', response.json()['status'])

    def getQuickTestResultPath(self, quickTestHandle):
        """
        quickTestHandle = /api/v1/sessions/1/ixnetwork/quickTest/rfc2544throughput/2
        """
        response = self.ixnObj.get(self.ixnObj.httpHeader + quickTestHandle + '/results')
        # "resultPath": "C:\\Users\\hgee\\AppData\\Local\\Ixia\\IxNetwork\\data\\result\\DP.Rfc2544Tput\\10694b39-6a8a-4e70-b1cd-52ec756910c3\\Run0001"
        return response.json()['resultPath']

    def getQuickTestResult(self, quickTestHandle, attribute):
        """
        Description
            Get Quick Test result attributes

        Parameter
            quickTestHandle: The Quick Test object handle

        attribute options to get:
           result - Returns pass
           status - Returns none
           progress - blank or Trial 1/1 Iteration 1, Size 64, Rate 10 % Wait for 2 seconds Wait 70.5169449%complete
           startTime - Returns 04/21/17 14:35:42
           currentActions
           waitingStatus
           resultPath
           isRunning - Returns True or False
           trafficStatus
           duration - Returns 00:01:03
           currentViews
        """
        response = self.ixnObj.get(quickTestHandle+'/results')
        return response.json()[attribute]

    def getQuickTestCsvFiles(self, quickTestHandle, copyToPath, csvFile='all'):
        """
        Description
            Copy Quick Test CSV result files to a specified path on either Windows or Linux.
            Note: Currently only supports copying from Windows.
                  Copy from Linux is coming in November.

        quickTestHandle: The Quick Test handle.
        copyToPath: The destination path to copy to.
                    If copy to Windows: c:\\Results\\Path
                    If copy to Linux: /home/user1/results/path

        csvFile: A list of CSV files to get: 'all', one or more CSV files to get:
                 AggregateResults.csv, iteration.csv, results.csv, logFile.txt, portMap.csv
        """
        resultsPath = self.getQuickTestResultPath(quickTestHandle)
        self.ixnObj.logInfo('\ngetQuickTestCsvFiles: %s' % resultsPath)
        if csvFile == 'all':
            getCsvFiles = ['AggregateResults.csv', 'iteration.csv', 'results.csv', 'logFile.txt', 'portMap.csv']
        else:
            if type(csvFile) is not list:
                getCsvFiles = [csvFile]
            else:
                getCsvFiles = csvFile

        for eachCsvFile in getCsvFiles:
            # Backslash indicates the results resides on a Windows OS.
            if '\\' in resultsPath:
                if bool(re.match('[a-z]:.*', copyToPath, re.I)):
                    self.fileMgmtObj.copyFileWindowsToLocalWindows(resultsPath+'\\{0}'.format(eachCsvFile), copyToPath)
                else:
                    self.fileMgmtObj.copyFileWindowsToLocalLinux(resultsPath+'\\{0}'.format(eachCsvFile), copyToPath)
            else:
                # TODO: Copy from Linux to Windows and Linux to Linux.
                pass

    def getQuickTestPdf(self, quickTestHandle, copyToLocalPath, where='remoteLinux', renameDestinationFile=None, includeTimestamp=False):
        """
        Description
           Generate Quick Test result to PDF and retrieve the PDF result file.

        Parameter
           where: localWindows|remoteWindows|remoteLinux. The destination.
           copyToLocalPath: The local destination path to store the PDF result file.
           renameDestinationFile: Rename the PDF file.
           includeTimestamp: True|False.  Set to True if you don't want to overwrite previous result file.
        """
        response = self.ixnObj.post(self.ixnObj.httpHeader+quickTestHandle+'/operations/generateReport', data={'arg1': quickTestHandle})
        if response.json()['url'] != '':
            if self.ixnObj.waitForComplete(response, self.ixnObj.httpHeader+response.json()['url']) == 1:
                raise IxNetRestApiException

            if where == 'localWindows':
                response = self.ixnObj.get(self.ixnObj.httpHeader+response.json()['url'])
                self.fileMgmtObj.copyFileWindowsToLocalWindows(response.json()['result'], copyToLocalPath, renameDestinationFile, includeTimestamp)
            if where == 'remoteWindows':
                # TODO: Work in progress.  Not sure if this is possible.
                resultPath = self.getQuickTestResultPath(quickTestHandle)
                #self.ixnObj.copyFileWindowsToRemoteWindows(response.json()['result'], copyToLocalPath, renameDestinationFile, includeTimestamp)
                self.fileMgmtObj.copyFileWindowsToRemoteWindows(resultPath, copyToLocalPath, renameDestinationFile, includeTimestamp)
            if where == 'remoteLinux':
                linuxResultPath = self.getQuickTestResultPath(quickTestHandle)
                self.fileMgmtObj.copyFileWindowsToLocalLinux(linuxResultPath+'\\TestReport.pdf', copyToLocalPath, renameDestinationFile, includeTimestamp)
        else:
            self.ixnObj.logInfo('\ngetQuickTestPdf failed. Result path = %s' % response.json()['result'])

    def runQuickTest(self, quickTestName, timeout=90):
        """
        Description
            Run the Quick test

        Parameter
            quickTestName: <str>: name of the quick test to run
            timeout: <int>: duration for quick test to run. Default=90 seconds

        Example
            runQuickTest("Macro_17_57_14_294", timeout=180)

        Return

        Note: operation run will keep checking the status of execution for the specified timeout
        """
        eventSchedulerHandle = self.getQuickTestHandleByName(quickTestName)

        url = self.ixnObj.sessionUrl + '/quickTest/eventScheduler/operations/run'
        data = {"arg1": eventSchedulerHandle}
        response = self.ixnObj.post(url, data=data)
        self.ixnObj.waitForComplete(response, url + '/' + response.json()['id'], timeout=timeout)

    def deleteQuickTest(self, quickTestName):
        """
        Description
            Delete the  Quick test.

        Parameter
            quickTestName: <str>: name of the quick test to delete

        Example
            deleteQuickTest("Macro_17_57_14_294")

        Return
        """
        eventSchedulerHandle = self.getQuickTestHandleByName(quickTestName)

        #delete the quick test
        url = self.ixnObj.httpHeader + eventSchedulerHandle
        self.ixnObj.delete(url)

    def configQuickTest(self, quickTestName, numOfTrials=1):
        """
        Description
            Configure a quick test

        Parameter
            quickTestName: <str>: name of the quick test to configure
            numOfTrials: <int>: number of iterations to run the quick test, default is 1

        Example
            configQuickTest("Macro_17_57_14_294")
            configQuickTest("Macro_17_57_14_294", numOfTrials=2)

        Return
            event scheduler handle on success or exception on failure
        """
        url = self.ixnObj.sessionUrl + '/quickTest/eventScheduler'
        response = self.ixnObj.post(url, data={"forceApplyQTConfig": "true", "mode": "existingMode", "name": quickTestName})

        eventSchedulerHandle = response.json()["links"][0]["href"]
        url = self.ixnObj.httpHeader + eventSchedulerHandle + '/eventScheduler'
        self.ixnObj.post(url, data={"enabled": "true", "itemId": quickTestName, "itemName": quickTestName})

        url = self.ixnObj.httpHeader + eventSchedulerHandle + '/testConfig'
        self.ixnObj.patch(url, data={"numTrials": numOfTrials})

        return eventSchedulerHandle
